#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import calendar
import datetime
import json
import logging
import os
from functools import reduce
from typing import Set, Dict, List, Iterable, Any, Tuple, Union, Optional
from urllib.parse import urlparse, urlunparse

from .base import dump_report_data, user_choice, field_to_title, report_output_parser
from .enterprise import EnterpriseCommand
from .. import api, crypto, utils, loginv3, constants
from ..display import bcolors
from ..error import CommandError
from ..params import KeeperParams
from ..proto import enterprise_pb2, BI_pb2, APIRequest_pb2


def register_commands(commands):
    commands['msp-down'] = GetMSPDataCommand()
    commands['msp-info'] = MSPInfoCommand()
    commands['msp-add'] = MSPAddCommand()
    commands['msp-remove'] = MSPRemoveCommand()
    commands['msp-update'] = MSPUpdateCommand()
    commands['msp-legacy-report'] = MSPLegacyReportCommand()
    commands['msp-billing-report'] = MSPBillingReportCommand()
    commands['msp-convert-node'] = MSPConvertNodeCommand()
    commands['msp-copy-role'] = MSPCopyRoleCommand()
    commands['switch-to-mc'] = SwitchToMcCommand()


def register_command_info(aliases, command_info):
    aliases['md'] = 'msp-down'
    aliases['mi'] = 'msp-info'
    aliases['ma'] = 'msp-add'
    aliases['mrm'] = 'msp-remove'
    aliases['mu'] = 'msp-update'
    aliases['mlr'] = 'msp-legacy-report'
    aliases['mbr'] = 'msp-billing-report'

    for p in [msp_down_parser, msp_info_parser, msp_add_parser, msp_remove_parser, msp_update_parser,
              msp_copy_role_parser, msp_legacy_report_parser, msp_billing_report_parser, switch_to_mc_parser]:
        command_info[p.prog] = p.description


msp_down_parser = argparse.ArgumentParser(prog='msp-down', usage='msp-down',
                                          description='Download current MSP data from the Keeper Cloud')

msp_info_parser = argparse.ArgumentParser(prog='msp-info', usage='msp-info', parents=[report_output_parser],
                                          description='Displays MSP details, such as managed companies and pricing')
msp_info_parser.add_argument('-p', '--pricing', dest='pricing', action='store_true', help='Display pricing information')
msp_info_parser.add_argument('-r', '--restriction', dest='restriction', action='store_true',
                             help='Display MSP restriction information')
msp_info_parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', help='Print details')
# msp_info_parser.add_argument('-u', '--users', dest='users', action='store_true', help='print user list')

msp_update_parser = argparse.ArgumentParser(prog='msp-update', usage='msp-update',
                                            description='Modify a Managed Company license')
msp_update_parser.add_argument('--node', dest='node', action='store', help='node name or node ID')
msp_update_parser.add_argument('-p', '--plan', dest='plan', action='store',
                               choices=[x[1] for x in constants.MSP_PLANS],
                               help=f'License plan: {", ".join((x[1] for x in constants.MSP_PLANS))}')
msp_update_parser.add_argument('-s', '--seats', dest='seats', action='store', type=int,
                               help='Maximum licences allowed. -1: unlimited')
msp_update_parser.add_argument('-f', '--file-plan', dest='file_plan', action='store',
                               help=f'File storage plan: {", ".join((x[2].lower() for x in constants.MSP_FILE_PLANS))}')
msp_update_parser.add_argument('-aa', '--add-addon', dest='add_addon', action='append', metavar='ADDON[:SEATS]',
                               help=f'Add add-ons: {", ".join(((x[0]+(":N" if x[2] else "")) for x in constants.MSP_ADDONS))}')
msp_update_parser.add_argument('-ra', '--remove-addon', dest='remove_addon', action='append', metavar='ADDON',
                               help=f'Add add-ons: {", ".join((x[0] for x in constants.MSP_ADDONS))}')
msp_update_parser.add_argument('mc', action='store',
                               help='Managed Company identifier (name or id). Ex. 3862 OR "Keeper Security, Inc."')

ranges = ['today', 'yesterday', 'last_7_days', 'last_30_days', 'month_to_date', 'last_month', 'year_to_date', 'last_year']
msp_legacy_report_parser = argparse.ArgumentParser(prog='msp-legacy-report', parents=[report_output_parser],
                                                   description='Generate MSP legacy billing report')
group = msp_legacy_report_parser.add_argument_group('Pre-defined date ranges')
group.add_argument('--range', dest='range', choices=ranges, default='last_30_days',
                   help="Pre-defined data ranges to run the report.")
group = msp_legacy_report_parser.add_argument_group('Custom date ranges')
group.add_argument('--from', dest='from_date',
                   help='Run report from this date. Value in ISO 8601 format (YYYY-mm-dd) or Unix timestamp format. '
                        'Only applicable to the `audit` report AND when there is no `range` specified. '
                        'Example: `2020-08-18` or `1596265200`')
group.add_argument('--to', dest='to_date',
                   help='Run report until this date. Value in ISO 8601 format (YYYY-mm-dd) '
                        'or Unix timestamp format. Only applicable to the `audit` report AND '
                        'when there is no `range` specified.'
                        'Example: `2020-08-18` or `1596265200`')

msp_billing_report_parser = argparse.ArgumentParser(prog='msp-billing-report', parents=[report_output_parser],
                                                    description='Generate MSP billing reports')
msp_billing_report_parser.add_argument('--month', dest='month', action='store', metavar='YYYY-MM', help='Month for billing report: 2022-02')
msp_billing_report_parser.add_argument('-d', '--show-date', dest='show_date', action='store_true', help='Breakdown report by date')
msp_billing_report_parser.add_argument('-c', '--show-company', dest='show_company', action='store_true', help='Breakdown report by managed company')


msp_add_parser = argparse.ArgumentParser(prog='msp-add', description='Add a Managed Company to the tenant')
msp_add_parser.add_argument('--node', dest='node', action='store', help='node name or node ID')
msp_add_parser.add_argument('-s', '--seats', dest='seats', action='store', type=int,
                            help='Maximum licences allowed. -1: unlimited')
msp_add_parser.add_argument('-p', '--plan', dest='plan', action='store', required=True,
                            choices=[x[1] for x in constants.MSP_PLANS],
                            help=f'License plan: {", ".join((x[1] for x in constants.MSP_PLANS))}')
msp_add_parser.add_argument('-f', '--file-plan', dest='file_plan', action='store',
                            help=f'File storage plan: {", ".join((x[2].lower() for x in constants.MSP_FILE_PLANS))}')
msp_add_parser.add_argument('-a', '--addon', dest='addon', action='append', metavar='ADDON[:SEATS]',
                            help=f'Add-ons: {", ".join((x[0]+(":N" if x[2] else "") for x in constants.MSP_ADDONS))}')
msp_add_parser.add_argument('name', action='store', help='Managed Company name')

msp_remove_parser = argparse.ArgumentParser(prog='msp-remove', description='Remove a managed company (MC) tenant')
msp_remove_parser.add_argument('-f', '--force', dest='force', action='store_true',
                               help='do not prompt for confirmation')
msp_remove_parser.add_argument('mc', action='store',
                               help='Managed Company identifier (name or id). Ex. 3862 OR "Keeper Security, Inc."')

msp_convert_node_parser = argparse.ArgumentParser(prog='msp-convert-node',
                                                  description='Converts MSP node into Managed Company.')
msp_convert_node_parser.add_argument('-s', '--seats', dest='seats', action='store', type=int,
                                     help='Number of seats')
msp_convert_node_parser.add_argument('-p', '--plan', dest='plan', action='store',
                                     choices=[x[1] for x in constants.MSP_PLANS],
                                     help='License Plan')
msp_convert_node_parser.add_argument('node', action='store', help='node name or node ID')

msp_copy_role_parser = argparse.ArgumentParser(
    prog='msp-copy-role', description='Copy role with enforcements to a Managed Company')
msp_copy_role_parser.add_argument('-r', '--role', dest='role', action='append',
                                  help='Role Name or ID. Can be repeated.')
msp_copy_role_parser.add_argument(
    'mc', action='store', nargs='+', help='Managed Company identifier (name or id)."')

switch_to_mc_parser = argparse.ArgumentParser(prog='switch-to-mc',
                                              description='Switch user\'s context to a Managed Company')
switch_to_mc_parser.add_argument('mc', action='store',
                                 help='Managed Company identifier (name or id). Ex. 3862 OR "Keeper Security, Inc."')
switch_to_msp_parser = argparse.ArgumentParser(prog='switch-to-msp',
                                               description='Switch user\'s context back to the MSP tenant')


msp_params = None
mc_params_dict = {}
current_mc_id = None


def bi_url(params, endpoint):
    p = urlparse(params.rest_context.server_base)
    return urlunparse((p.scheme, p.netloc, '/bi_api/v2/enterprise_console/' + endpoint, None, None, None))


class GetMSPDataCommand(EnterpriseCommand):

    def get_parser(self):
        return msp_down_parser

    def execute(self, params, **kwargs):
        api.query_enterprise(params)


class MSPMixin:
    @staticmethod
    def price_text_short(price_info):
        price = ''
        if isinstance(price_info, dict):
            if 'amount' in price_info:
                currency = price_info.get('currency')
                if currency:
                    if currency == 'USD':
                        currency = '$'
                    elif currency == 'EUR':
                        currency = '\u20AC'
                    elif currency == 'GBP':
                        currency = '\u00a3'
                    elif currency == 'JPY':
                        currency = '\u00a5'
                    price += currency

                price += str(price_info['amount'])
        return price

    @staticmethod
    def price_text(price_info):
        price = MSPMixin.price_text_short(price_info)
        if price and isinstance(price_info, dict):
            unit = price_info.get('unit')
            if unit:
                if unit == 'USER_MONTH':
                    unit = 'user/month'
                elif unit == 'MONTH':
                    unit = 'month'
                elif unit == 'USER_CONSUMED_MONTH':
                    unit = '50k API calls/month'
                price += '/' + unit
        return price

    @staticmethod
    def get_msp_addons(params):   # type: (Any) -> Dict[int, str]
        if 'msp_addons' not in params.enterprise:
            url = bi_url(params, 'mapping/addons')
            rq = BI_pb2.MappingAddonsRequest()
            rs = api.communicate_rest(params, rq, url, rs_type=BI_pb2.MappingAddonsResponse)
            addon_map = {x.id: x.name for x in rs.addons}
            params.enterprise['msp_addons'] = addon_map
        return params.enterprise['msp_addons']

    @staticmethod
    def get_msp_pricing(params):
        if 'msp_pricing' not in params.enterprise:
            plan_map = {x[0]: x[1] for x in constants.MSP_PLANS}
            file_map = {x[0]: x[1] for x in constants.MSP_FILE_PLANS}
            addon_map = MSPMixin.get_msp_addons(params)

            pricing = {}
            params.enterprise['msp_pricing'] = pricing

            url = bi_url(params, 'subscription/mc_pricing')
            rq = BI_pb2.SubscriptionMcPricingRequest()
            rs = api.communicate_rest(params, rq, url, rs_type=BI_pb2.SubscriptionMcPricingResponse)

            units = BI_pb2.Cost.AmountPer.keys()
            currencies = BI_pb2.Currency.keys()
            pricing['mc_base_plans'] = {}
            for p in rs.basePlans:
                if p.id in plan_map:
                    pricing['mc_base_plans'][plan_map[p.id]] = {
                        'amount': p.cost.amount,
                        'unit': units[p.cost.amountPer],
                        'currency': currencies[p.cost.currency]
                    }
            pricing['mc_addons'] = {}
            for p in rs.addons:
                if p.id in addon_map:
                    pricing['mc_addons'][addon_map[p.id]] = {
                        'amount': p.cost.amount,
                        'unit': units[p.cost.amountPer],
                        'currency': currencies[p.cost.currency]
                    }
            pricing['mc_file_plans'] = {}
            for p in rs.filePlans:
                if p.id in file_map:
                    pricing['mc_file_plans'][file_map[p.id]] = {
                        'amount': p.cost.amount,
                        'unit': units[p.cost.amountPer],
                        'currency': currencies[p.cost.currency]
                    }

        return params.enterprise['msp_pricing']


class SwitchToMspCommand(EnterpriseCommand):
    def get_parser(self):
        return switch_to_msp_parser

    def execute(self, params, **kwargs):
        global current_mc_id, msp_params
        if current_mc_id is None:
            raise CommandError('switch-to-msp', "Already MSP")

        current_mc_id = None
        if msp_params is not None:
            api.query_enterprise(msp_params)
        logging.info("Switched back to MSP")


class SwitchToMcCommand(EnterpriseCommand, MSPMixin):
    def get_parser(self):
        return switch_to_mc_parser

    def execute(self, params, **kwargs):
        global current_mc_id, msp_params
        if current_mc_id is not None:
            raise CommandError('switch-to-mc',
                               f"Already switched to Managed Company id={current_mc_id}")

        managed_companies = params.enterprise['managed_companies']
        mc_input = kwargs.get('mc')
        current_mc = get_mc_by_name_or_id(managed_companies, mc_input)
        if not current_mc:
            raise CommandError('switch-to-mc', f'Managed Company \"{mc_input}\" not found')

        current_mc_id = current_mc['mc_enterprise_id']

        if current_mc_id not in mc_params_dict:
            mc_params = api.login_and_get_mc_params_login_v3(params, current_mc_id)
            mc_params_dict[current_mc_id] = mc_params
        else:
            mc_params = mc_params_dict[current_mc_id]
            api.query_enterprise(mc_params)

        msp_params = params
        logging.info("Switched to MC '%s'", current_mc['mc_enterprise_name'])


class MSPInfoCommand(EnterpriseCommand, MSPMixin):
    def get_parser(self):
        return msp_info_parser

    def execute(self, params, **kwargs):
        report_format = kwargs.get('format')
        output_file = kwargs.get('output')
        if kwargs.get('restriction'):
            permits = next((x['msp_permits'] for x in params.enterprise.get('licenses', []) if 'msp_permits' in x), None)
            if permits:
                all_products = {x[1].lower(): x[2] for x in constants.MSP_PLANS}
                all_addons = {x[0].lower(): x[3] for x in constants.MSP_ADDONS}
                all_file_plans = {x[1].lower(): x[2] for x in constants.MSP_FILE_PLANS}
                max_file_plan = permits['max_file_plan_type']
                table = [
                    ['Allow Unlimited Licenses', permits['allow_unlimited_licenses']],
                    ['Allowed Products', [x + f' ({all_products.get(x.lower(), "")})' for x in permits['allowed_mc_products']]],
                    ['Allowed Add-Ons', [x + f' ({all_addons.get(x.lower(), "")})' for x in permits['allowed_add_ons']]],
                    ['Max File Storage plan', all_file_plans.get(max_file_plan.lower(), max_file_plan)]
                ]
                return dump_report_data(table, ['Permit Name', 'Value'], fmt=report_format, filename=output_file)
            else:
                logging.info('MSP has no restrictions')
            return

        if kwargs.get('pricing'):
            pricing = MSPMixin.get_msp_pricing(params)

            header = ['category', 'name', 'code', 'price']
            table = []
            if 'mc_base_plans' in pricing:
                plans = pricing['mc_base_plans']
                for plan in constants.MSP_PLANS:
                    code = plan[1]
                    if code in plans:
                        info = plans[code]
                        row = ['Product', plan[2], code, MSPMixin.price_text(info)]
                        table.append(row)
            if 'mc_addons' in pricing:
                addons = pricing['mc_addons']
                for addon in constants.MSP_ADDONS:
                    code = addon[0]
                    if code in addons:
                        info = addons[code]
                        row = ['Addon', addon[1], code, MSPMixin.price_text(info)]
                        table.append(row)

            if 'mc_file_plans' in pricing:
                plans = pricing['mc_file_plans']
                for addon in constants.MSP_FILE_PLANS:
                    plan = addon[1]
                    if plan in plans:
                        info = plans[plan]
                        row = ['File Plan', addon[2], plan, MSPMixin.price_text(info)]
                        table.append(row)
            if report_format != 'json':
                header = [field_to_title(x) for x in header]
            return dump_report_data(table, header, fmt=report_format, filename=output_file)

        if 'managed_companies' in params.enterprise:
            sort_dict = {x[0]: i for i, x in enumerate(constants.MSP_ADDONS)}
            verbose = kwargs.get('verbose')
            header = ['company_id', 'company_name', 'node', 'plan', 'storage', 'addons', 'allocated', 'active']
            table = []
            plan_map = {x[1]: x[2] for x in constants.MSP_PLANS}
            file_plan_map = {x[1]: x[2] for x in constants.MSP_FILE_PLANS}
            for mc in params.enterprise['managed_companies']:
                node_id = mc['msp_node_id']
                if verbose:
                    node_path = str(node_id)
                else:
                    node_path = self.get_node_path(params, node_id, False)
                file_plan = mc['file_plan_type']
                file_plan = file_plan_map.get(file_plan, file_plan)
                addons = [x['name'] for x in mc.get('add_ons', [])]
                addons.sort(key=lambda x: sort_dict.get(x, -1))
                if not verbose:
                    addons = len(addons)
                plan = mc['product_id']
                if not verbose:
                    plan = plan_map.get(plan, plan)
                seats = mc['number_of_seats']
                if seats > 2000000:
                    seats = None
                table.append([mc['mc_enterprise_id'], mc['mc_enterprise_name'], node_path,
                              plan, file_plan, addons, seats, mc['number_of_users']])
            table.sort(key=lambda x: x[1].lower())
            if report_format != 'json':
                header = [field_to_title(x) for x in header]
            return dump_report_data(table, header, row_number=True, fmt=report_format, filename=output_file)
        else:
            logging.info("No Managed Companies")


class MSPUpdateCommand(EnterpriseCommand):
    def get_parser(self):
        return msp_update_parser

    def execute(self, params, **kwargs):
        managed_companies = params.enterprise['managed_companies']
        mc_input = kwargs.get('mc')
        current_mc = get_mc_by_name_or_id(managed_companies, mc_input)
        if not current_mc:
            raise CommandError('msp-remove', f'Managed Company \"{mc_input}\" not found')

        rq = {
            'command': 'enterprise_update_by_msp',
            'enterprise_id': current_mc['mc_enterprise_id'],
            'enterprise_name': current_mc['mc_enterprise_name'],
            'product_id': current_mc['product_id'],
            'seats': current_mc['number_of_seats'],
        }

        node_name = kwargs.get('node')
        if node_name:
            nodes = list(self.resolve_nodes(params, node_name))
            if len(nodes) == 0:
                raise CommandError('msp-update', f'Node \"{node_name}\" is not found')
            if len(nodes) > 1:
                raise CommandError('msp-update', f'More than one nodes \"{node_name}\" are found')
            rq['node_id'] = nodes[0]['node_id']

        permits = next((x['msp_permits'] for x in params.enterprise.get('licenses', []) if 'msp_permits' in x), None)

        plan_name = kwargs.get('plan')
        if plan_name:
            plan_name = plan_name.lower()
            product_plan = next((x for x in constants.MSP_PLANS if x[1].lower() == plan_name), None)
            if not product_plan:
                raise CommandError('msp-update', f'Managed Company plan \"{plan_name}\" is not found')
            if permits:
                has_plan = any((True for x in permits['allowed_mc_products'] if x.lower() == plan_name))
                if not has_plan:
                    raise CommandError('msp-update', f'Managed Company plan \"{plan_name}\" is not allowed')
            rq['product_id'] = product_plan[1]

        seats = kwargs.get('seats')
        if isinstance(seats, int):
            if seats < 0 and permits:
                if permits['allow_unlimited_licenses'] is False:
                    raise CommandError('msp-update', 'Managed Company unlimited licences are not allowed')
            rq['seats'] = seats if seats >= 0 else 2147483647

        plan_name = kwargs.get('file_plan')
        if plan_name:
            plan_name = plan_name.lower()
            file_plan = next((x for x in constants.MSP_FILE_PLANS if plan_name in (y.lower() for y in x if isinstance(y, str))), None)
            if not file_plan:
                raise CommandError('msp-update', f'File plan \"{plan_name}\" is not found')
            if permits:
                allowed_file_plan_name = permits['max_file_plan_type'].lower()
                if allowed_file_plan_name:
                    allowed_plan = next((x for x in constants.MSP_FILE_PLANS if allowed_file_plan_name == x[1].lower()), None)
                    if allowed_plan and allowed_plan[0] < file_plan[0]:
                        raise CommandError('msp-update',
                                           f'Managed Company file storage \"{file_plan[2]}\" is not allowed')
            product_id = rq['product_id'].lower()
            product_plan = next((x for x in constants.MSP_PLANS if product_id == x[1].lower()), None)
            if product_plan and product_plan[3] < file_plan[0]:
                rq['file_plan_type'] = file_plan[1]

        addons = {}
        for ao in current_mc.get('add_ons', []):
            if not ao['enabled']:
                continue
            if ao.get('included_in_product') is True:
                continue
            addon_name = ao['name']
            keep_addon = {
                'add_on': addon_name
            }
            seats = ao.get('seats')
            if seats > 0:
                keep_addon['seats'] = seats
            addons[addon_name] = keep_addon

        for action in ('add_addon', 'remove_addon'):
            action_addons = kwargs.get(action)
            if isinstance(action_addons, list):
                for aon in action_addons:
                    addon_name, sep, seats = aon.partition(':')
                    addon_name = addon_name.lower()
                    addon = next((x for x in constants.MSP_ADDONS if addon_name == x[0]), None)
                    if addon is None:
                        raise CommandError('msp-update',f'Addon \"{addon_name}\" is not found')
                    addon_seats = 0
                    if sep == ':' and addon[2] and action == 'add_addon':
                        try:
                            addon_seats = int(seats)
                        except:
                            raise CommandError('msp-update',
                                               f'Addon \"{addon_name}\". Number of seats \"{seats}\" is not integer')
                    if action == 'add_addon':
                        if permits:
                            if addon_name not in (x.lower() for x in permits['allowed_add_ons']):
                                raise CommandError('msp-update',
                                                   f'Managed Company add-on \"{addon_name}\" is not allowed')
                        add_addon = {
                            'add_on': addon_name
                        }
                        if addon_seats > 0:
                            add_addon['seats'] = seats
                        addons[addon_name] = add_addon
                    else:
                        if addon_name in addons:
                            del addons[addon_name]
        rq['add_ons'] = list(addons.values())
        rs = api.communicate(params, rq)
        if rs['result'] == 'success':
            mc_from_rs = find(lambda mc: mc['mc_enterprise_id'] == rs["enterprise_id"], managed_companies)
            logging.info("Successfully updated '%s' id=%d", mc_from_rs['mc_enterprise_name'], mc_from_rs['mc_enterprise_id'])
            api.query_enterprise(params)


class DailySnapshot(object):
    def __init__(self, mc_id, date_no):
        self.mc_enterprise_id = mc_id
        self.date_no = date_no

    def __eq__(self, other):
        if isinstance(other, DailySnapshot):
            return self.mc_enterprise_id == other.mc_enterprise_id and self.date_no == other.date_no
        return False

    def __str__(self):
        return f'MC ID: {self.mc_enterprise_id}; Date: {self.date_no}'

    def __hash__(self):
        b = (self.mc_enterprise_id << 32) + self.date_no
        return b.__hash__()

    @staticmethod
    def merge_units(unit_iter):
        # type: (Iterable[Dict[int, Union[int, Tuple[int, int]]]]) -> Dict[int, Tuple[int, int]]
        ret = {}   # type: Dict[int, Tuple[int, int]]
        for units in unit_iter:
            if not isinstance(units, dict):
                continue
            for unit in units:
                if not isinstance(unit, int):
                    continue
                count = units[unit]
                if not isinstance(count, (int, tuple)):
                    continue
                if isinstance(count, int):
                    qty = count
                    days = 1
                else:
                    qty = count[0]
                    days = count[1]
                if unit in ret:
                    q1, d1 = ret[unit]
                    ret[unit] = (q1 + qty, d1 + days)
                else:
                    ret[unit] = (qty, days)
        return ret


class MSPBillingReportCommand(EnterpriseCommand):
    LAST_USER = ''
    SNAPSHOT_CACHE = {}  # type: Dict[str, Dict[DailySnapshot, Dict[int, int]]]
    COMPANY_CACHE = {}

    def get_parser(self):
        return msp_billing_report_parser

    @staticmethod
    def is_plan_id(msp_id):   # type: (int) -> bool
        return 0 < msp_id < 100

    @staticmethod
    def is_storage_plan_id(msp_id):   # type: (int) -> bool
        return 100 < msp_id < 10000

    @staticmethod
    def is_addon_id(msp_id):   # type: (int) -> bool
        return 10000 < msp_id

    @staticmethod
    def get_count_id(msp_id):
        if 0 < msp_id < 100:
            return msp_id
        if 100 < msp_id < 10000:
            return msp_id // 100
        if msp_id > 10000:
            return msp_id // 10000
        return 0

    @staticmethod
    def get_daily_snapshots(params, year, month):
        if MSPBillingReportCommand.LAST_USER:
            if MSPBillingReportCommand.LAST_USER != params.user:
                MSPBillingReportCommand.SNAPSHOT_CACHE.clear()
                MSPBillingReportCommand.COMPANY_CACHE.clear()
        MSPBillingReportCommand.LAST_USER = params.user

        key = f'{year}-{month}'
        if key not in MSPBillingReportCommand.SNAPSHOT_CACHE:
            rq = BI_pb2.ReportingDailySnapshotRequest()
            rq.year = year
            rq.month = month
            url = bi_url(params, 'reporting/daily_snapshot')
            rs = api.communicate_rest(params, rq, url, rs_type=BI_pb2.ReportingDailySnapshotResponse)
            for company in rs.mcEnterprises:
                MSPBillingReportCommand.COMPANY_CACHE[company.id] = company.name

            snapshot = {}
            for record in rs.records:
                units = {}
                if record.maxLicenseCount > 0:
                    if record.maxBasePlanId > 0:
                        units[record.maxBasePlanId] = record.maxLicenseCount
                    if record.maxFilePlanTypeId > 0:
                        units[record.maxFilePlanTypeId * 100] = record.maxLicenseCount
                    for addon in record.addons:
                        if addon.maxAddonId > 0:
                            units[addon.maxAddonId * 10000] = addon.units
                mc_id = record.mcEnterpriseId
                ds = datetime.datetime.fromtimestamp(record.date // 1000, tz=datetime.timezone.utc)
                dt = ds.date()
                daily = DailySnapshot(mc_id, dt.toordinal())
                snapshot[daily] = units
            MSPBillingReportCommand.SNAPSHOT_CACHE[key] = snapshot
        return MSPBillingReportCommand.SNAPSHOT_CACHE[key]

    @staticmethod
    def get_max_product_count(period_snapshots, product, mc_id=None):
        # type: (Dict[DailySnapshot, Dict[int, int]], int, Optional[str]) -> int
        daily_counts = dict()
        for ds, counts in period_snapshots.items():
            if mc_id is not None and mc_id != ds.mc_enterprise_id:
                continue
            date = ds.date_no
            counts_group = daily_counts.get(date, [])
            counts_group.append(counts.get(product) or 0)
            daily_counts[date] = counts_group
        daily_totals = [sum(counts) for counts in daily_counts.values()]
        return max(daily_totals)

    @staticmethod
    def get_bounding_snapshots(period_snapshots, mc_id=None):
        # type: (Dict[DailySnapshot, Dict[int, int]], Optional[str]) -> Tuple[Union[Dict[int, int], Dict[int, Tuple[int, int]]], ...]
        snap_keys = period_snapshots.keys()

        def get_mc_bounding_snapshots(mc):
            # type: (str) -> Tuple[Optional[Dict[int, int]], ...]
            mc_snapshots = {ds for ds in snap_keys if ds.mc_enterprise_id == mc}
            if not len(mc_snapshots):
                return None, None, None

            dates = {ds.date_no for ds in mc_snapshots}
            start_date, end_date = min(dates), max(dates)
            start_ds = next((ds for ds in mc_snapshots if ds.date_no == start_date), None)
            end_ds = next((ds for ds in mc_snapshots if ds.date_no == end_date), None)
            return period_snapshots.get(start_ds), period_snapshots.get(end_ds)

        if mc_id is None:
            start_snapshots = []   # type: List[Dict[int, int]]
            end_snapshots = []     # type: List[Dict[int, int]]
            mc_ids = {ds.mc_enterprise_id for ds in snap_keys}
            # Get the first and last snapshot for each MC
            for mc in mc_ids:
                start, end = get_mc_bounding_snapshots(mc)
                start and start_snapshots.append(start)
                end and end_snapshots.append(end)

            # Merge all first snapshots together, then all last snapshots
            merge_counts = lambda x, y: DailySnapshot.merge_units((x, y))
            start_snapshot = reduce(merge_counts, start_snapshots)
            end_snapshot = reduce(merge_counts, end_snapshots)
            return start_snapshot, end_snapshot
        else:
            return get_mc_bounding_snapshots(mc_id)

    @staticmethod
    def get_num_reported_days(period_snapshots): # type: (Dict[DailySnapshot, Dict[int, int]]) -> int
        dates = [ds.date_no for ds in period_snapshots]
        return max(dates) - min(dates) + 1 if dates else 30

    def execute(self, params, **kwargs):
        month_str = kwargs.get('month')
        if not month_str:
            dt = datetime.datetime.now()
            month = dt.month
            year = dt.year
            month -= 1
            if month < 1:
                month += 12
                year -= 1
        else:
            year_part, sep, month_part = month_str.partition('-')
            try:
                year = int(year_part)
                month = int(month_part)
            except:
                logging.warning('Given month \"%s\" is not valid. YYYY-MM', month_str)
                return
        daily_counts = MSPBillingReportCommand.get_daily_snapshots(params, year, month)
        title = f'Consumption Billing Statement: {calendar.month_name[month]} {year}'
        headers = []
        table = []

        show_date = kwargs.get('show_date', False)
        show_company = kwargs.get('show_company', False)
        merged_counts = {}  # type: Dict[DailySnapshot, Dict[int, Tuple[int, int]]]
        for dc in daily_counts:
            d = DailySnapshot(dc.mc_enterprise_id if show_company else 0, dc.date_no if show_date else 0)
            merged_counts[d] = DailySnapshot.merge_units((merged_counts.get(d), daily_counts[dc]))

        if show_date:
            headers.append('date')
        if show_company:
            headers.extend(('company', 'company_id'))
        headers.extend(('product', 'licenses', 'rate'))
        if not show_date:
            headers.extend(['avg_per_day', 'initial_licenses', 'final_licenses', 'max_licenses'])
        plan_lookup = {x[0]: x for x in constants.MSP_PLANS}
        storage_lookup = {x[0]: x for x in constants.MSP_FILE_PLANS}
        addon_lookup = {}
        addons = {x[0]: x for x in constants.MSP_ADDONS}
        for a_id, a_name in MSPMixin.get_msp_addons(params).items():
            if a_name in addons:
                addon_lookup[a_id] = addons[a_name]
        pricing = MSPMixin.get_msp_pricing(params)
        for point in merged_counts:
            day_str = str(datetime.date.fromordinal(point.date_no)) if show_date else ''
            company = MSPBillingReportCommand.COMPANY_CACHE.get(point.mc_enterprise_id, '') if show_company else ''
            start_snapshots, end_snapshots = (None, None) \
                if show_date \
                else MSPBillingReportCommand.get_bounding_snapshots(daily_counts, None if not show_company else point.mc_enterprise_id)
            counts = merged_counts[point]
            products = list(counts.keys())
            products.sort()
            for product in products:
                row = []
                if show_date:
                    row.append(day_str)
                if show_company:
                    row.extend((company, point.mc_enterprise_id))
                count_id = MSPBillingReportCommand.get_count_id(product)
                count, days = counts[product] \
                    if show_company \
                    else (counts[product][0], MSPBillingReportCommand.get_num_reported_days(daily_counts))

                product_name = ''
                rate_text = ''
                if MSPBillingReportCommand.is_plan_id(product):
                    plan = plan_lookup.get(count_id)
                    product_name = plan[2] if plan else str(count_id)
                    if plan and 'mc_base_plans' in pricing:
                        if plan[1] in pricing['mc_base_plans']:
                            rate = pricing['mc_base_plans'][plan[1]]
                            rate_text = MSPMixin.price_text_short(rate)
                elif MSPBillingReportCommand.is_storage_plan_id(product):
                    plan = storage_lookup.get(count_id)
                    product_name = plan[2] if plan else str(count_id)
                    if plan and 'mc_file_plans' in pricing:
                        if plan[1] in pricing['mc_file_plans']:
                            rate = pricing['mc_file_plans'][plan[1]]
                            rate_text = MSPMixin.price_text_short(rate)
                elif MSPBillingReportCommand.is_addon_id(product):
                    addon = addon_lookup.get(count_id)
                    product_name = addon[1] if addon else str(count_id)
                    if addon and 'mc_addons' in pricing:
                        if addon[0] in pricing['mc_addons']:
                            rate = pricing['mc_addons'][addon[0]]
                            rate_text = MSPMixin.price_text_short(rate)
                else:
                    product_name = str(product)

                row.extend((product_name, count, rate_text))
                if not show_date:
                    row.append(round(count / days, 2))
                    start_counts_data = 0 if start_snapshots is None else start_snapshots.get(product) or 0
                    end_counts_data = 0 if end_snapshots is None else end_snapshots.get(product) or 0
                    start_count = next(iter(start_counts_data)) if isinstance(start_counts_data, tuple) else start_counts_data
                    end_count = next(iter(end_counts_data)) if isinstance(end_counts_data, tuple) else end_counts_data
                    max_count = MSPBillingReportCommand.get_max_product_count(daily_counts, product, None if not show_company else point.mc_enterprise_id)
                    row.extend([start_count, end_count, max_count])

                table.append(row)

        output_format = kwargs.get('format')
        if output_format == 'table':
            headers = [field_to_title(x) for x in headers]
        return dump_report_data(table, headers, fmt=output_format, filename=kwargs.get('output'), title=title)


class MSPLegacyReportCommand(EnterpriseCommand):
    def get_parser(self):
        return msp_legacy_report_parser

    def execute(self, params, **kwargs):
        from_date_str = kwargs.get('from_date')
        to_date_str = kwargs.get('to_date')
        if not from_date_str or not to_date_str:
            # will use data range to query
            rng = kwargs.get('range')
            from_date, to_date = date_range_str_to_dates(rng)
        else:
            # will use start and end data
            if loginv3.CommonHelperMethods.check_int(from_date_str):
                from_date = datetime.datetime.fromtimestamp(int(from_date_str))
            else:
                from_date = datetime.datetime.strptime(from_date_str + " 00:00:00", "%Y-%m-%d %H:%M:%S")

            if loginv3.CommonHelperMethods.check_int(to_date_str):
                to_date = datetime.datetime.fromtimestamp(int(to_date_str))
            else:
                to_date = datetime.datetime.strptime(to_date_str + " 11:59:59", "%Y-%m-%d %H:%M:%S")

        from_date_timestamp = int(from_date.timestamp() * 1000)
        to_date_timestamp = int(to_date.timestamp() * 1000)

        rq = {
            'command': 'get_mc_license_adjustment_log',
            'from': from_date_timestamp,
            'to': to_date_timestamp
        }

        rs = api.communicate(params, rq)

        title = None
        headers = []
        table = []
        for log in rs['log']:
            table.append([log['id'], log['date'], log['enterprise_id'], log['enterprise_name'], log['status'],
                          log['new_number_of_seats'], log['new_product_type'], log['note'], log['price']])

        headers.extend(('id', 'time', 'company_id', 'company_name', 'status', 'number_of_allocations', 'plan', 'transaction_notes', 'price_estimate'))

        output_format = kwargs.get('format')
        if output_format == 'table':
            headers = [field_to_title(x) for x in headers]
        return dump_report_data(table, headers, fmt=output_format, filename=kwargs.get('output'), title=title)


class MSPAddCommand(EnterpriseCommand):
    def get_parser(self):
        return msp_add_parser

    def execute(self, params, **kwargs):
        node_id = None
        node_name = kwargs.get('node')
        if node_name:
            nodes = list(self.resolve_nodes(params, node_name))
            if len(nodes) == 0:
                logging.warning('Node \"%s\" is not found', node_name)
                return
            if len(nodes) > 1:
                logging.warning('More than one nodes \"%s\" are found', node_name)
                return
            node_id = nodes[0]['node_id']
        if node_id is None:
            root_nodes = list(self.get_user_root_nodes(params))
            if len(root_nodes) == 0:
                raise CommandError('msp-create', 'No root nodes were detected. Specify --node parameter')
            node_id = root_nodes[0]

        permits = next((x['msp_permits'] for x in params.enterprise.get('licenses', []) if 'msp_permits' in x), None)

        plan_name = kwargs.get('plan')
        if not plan_name and permits:
            allowed_products = permits['allowed_mc_products']
            if allowed_products:
                plan_name = allowed_products[0]
        if not plan_name:
            plan_name = constants.MSP_PLANS[0][1]
        if permits:
            has_plan = any((True for x in permits['allowed_mc_products'] if x.lower() == plan_name.lower()))
            if not has_plan:
                logging.warning('Managed Company plan \"%s\" is not allowed', plan_name)
                return

        plan_name = plan_name.lower()
        product_plan = next((x for x in constants.MSP_PLANS if x[1].lower() == plan_name), None)
        if not product_plan:
            logging.warning('Managed Company plan \"%s\" is not found', plan_name)
            return

        seats = kwargs.get('seats')
        if isinstance(seats, int) and seats < 0:
            if permits:
                if permits['allow_unlimited_licenses'] is False:
                    logging.warning('Managed Company unlimited licences are not allowed')
                    return
            seats = 2147483647

        name = kwargs['name']
        tree_key = utils.generate_aes_key()
        rq = {
            'command': 'enterprise_registration_by_msp',
            'node_id': node_id,
            'product_id': product_plan[1],
            'seats': seats if isinstance(seats, int) else 0,
            'enterprise_name': name,
            'encrypted_tree_key': utils.base64_url_encode(
                crypto.encrypt_aes_v2(tree_key, params.enterprise['unencrypted_tree_key'])),
            'role_data': utils.base64_url_encode(
                crypto.encrypt_aes_v1(json.dumps({'displayname': 'Keeper Administrator'}).encode(), tree_key)),
            'root_node': utils.base64_url_encode(
                crypto.encrypt_aes_v1(json.dumps({'displayname': 'root'}).encode(), tree_key))
        }

        plan_name = kwargs.get('file_plan')
        if plan_name:
            plan_name = plan_name.lower()
            file_plan = next((x for x in constants.MSP_FILE_PLANS if plan_name in (y.lower() for y in x if isinstance(y, str))), None)
            if not file_plan:
                logging.warning('File plan \"%s\" is not found', plan_name)
                return
            if product_plan[3] < file_plan[0]:
                rq['file_plan_type'] = file_plan[1]
            if permits:
                allowed_file_plan_name = permits['max_file_plan_type'].lower()
                if allowed_file_plan_name:
                    allowed_plan = next((x for x in constants.MSP_FILE_PLANS if allowed_file_plan_name == x[1].lower()), None)
                    if allowed_plan and allowed_plan[0] < file_plan[0]:
                        logging.warning('Managed Company file storage \"%s\" is not allowed', file_plan[2])
                        return

        addons = kwargs.get('addon')
        if isinstance(addons, list):
            rq['add_ons'] = []
            for v in addons:
                addon_name, sep, seats = v.partition(':')
                addon_name = addon_name.lower()
                addon = next((x for x in constants.MSP_ADDONS if x[0] == addon_name), None)
                if addon is None:
                    logging.warning('Addon \"%s\" is not found', addon_name)
                    return
                if permits:
                    if addon_name not in (x.lower() for x in permits['allowed_add_ons']):
                        logging.warning('Managed Company add-on \"%s\" is not allowed', addon_name)
                        return
                addon_seats = 0
                if sep == ':' and addon[2]:
                    try:
                        addon_seats = int(seats)
                    except:
                        logging.warning('Addon \"%s\". Number of seats \"%s\" is not integer', addon_name, seats)
                        return
                rqa = {
                    'add_on': addon[0]
                }
                if addon_seats > 0:
                    rqa['seats'] = addon_seats
                rq['add_ons'].append(rqa)

        company_id = -1
        rs = api.communicate(params, rq)
        if rs:
            company_id = rs.get('enterprise_id', -1)
            params.environment_variables['last_mc_id'] = str(company_id)
            logging.info('Managed company \"%s\" added. ID=%d', name, company_id)
        api.query_enterprise(params)
        return company_id


class MSPRemoveCommand(EnterpriseCommand):
    def get_parser(self):
        return msp_remove_parser

    def execute(self, params, **kwargs):
        mc_input = kwargs.get('mc', '')
        if not mc_input:
            raise CommandError('msp-remove', 'Managed Company name or id is required')
        managed_companies = params.enterprise.get('managed_companies', [])
        current_mc = get_mc_by_name_or_id(managed_companies, mc_input)
        if not current_mc:
            raise CommandError('msp-remove', f'Managed Company \"{mc_input}\" not found')
        if kwargs.get('force') is True:
            answer = 'y'
        else:
            answer = user_choice(bcolors.FAIL + bcolors.BOLD + 'ALERT!\n' + bcolors.ENDC + 'Remove Managed Company.\n\n' +
                                 'Removing will expire the licences for the managed company and your admin access for the account.\n' +
                                 f'Managed Company Name: \"{current_mc["mc_enterprise_name"]}\", Licences: {current_mc["number_of_seats"]}\n\n' +
                                 'I want to remove these licences managed vault folder and my access to the admin console from my MSP account.', 'yn', 'n')
        if answer.lower() == 'y':
            rq = {
                'command': 'enterprise_remove_by_msp',
                'enterprise_id': current_mc['mc_enterprise_id']
            }
            rs = api.communicate(params, rq)
            if rs:
                logging.info('Managed company \"%s\" removed. ID=%d', current_mc['mc_enterprise_name'], current_mc['mc_enterprise_id'])
            api.query_enterprise(params)


def get_mc_by_name_or_id(msc, name_or_id):
    if loginv3.CommonHelperMethods.check_int(name_or_id):
        # get by id
        found_mc = find(lambda mc: mc['mc_enterprise_id'] == int(name_or_id), msc)
    else:
        # get by company name (all lower case)
        found_mc = find(lambda mc: mc['mc_enterprise_name'].lower() == name_or_id.lower(), msc)

    return found_mc


def find(f, seq):
    """Return first item in sequence where f(item) == True."""
    for item in seq:
        if f(item):
            return item


def report_generation_message(filename, filetype):
    if filename:
        _, ext = os.path.splitext(filename)
        if not ext:
            filename += '.'+filetype

    return filename


def date_range_str_to_dates(range_str):

    if range_str not in ranges:
        raise CommandError('', "Given range %s is not supported. Supported ranges: %s" % (range_str, ranges))

    current_time = datetime.datetime.now()

    today_start_dt = current_time.replace(hour=0, minute=0, second=0)
    today_end_dt = current_time.replace(hour=11, minute=59, second=59)

    start_date = None
    end_date = None

    def last_day_of_month(dt):
        year = dt.strftime("%Y")                       # get the year
        month = str(int(dt.strftime("%m")) % 12 + 1)   # get month, watch rollover

        ldom = calendar.monthrange(int(year), int(month))[1]  # get num of days in this month

        last_date_of_month = dt.replace(hour=11, minute=59, second=59, day=ldom)

        return last_date_of_month

    if range_str == 'today':
        start_date = today_start_dt
        end_date = today_end_dt

    elif range_str == 'yesterday':
        start_date = today_start_dt - datetime.timedelta(1)
        end_date = today_end_dt - datetime.timedelta(1)

    elif range_str == 'last_7_days':
        start_date = today_start_dt - datetime.timedelta(7)
        end_date = today_end_dt

    elif range_str == 'last_30_days':
        start_date = today_start_dt - datetime.timedelta(30)
        end_date = today_end_dt

    elif range_str == 'month_to_date':
        start_date = today_start_dt.replace(hour=0, minute=0, second=0, day=1)
        end_date = today_end_dt

    elif range_str == 'last_month':
        last_month_num = current_time.month - 1 if current_time.month > 1 else 12
        last_month_dt = current_time.replace(month=last_month_num)

        start_date = current_time.replace(month=last_month_num, day=1, hour=0, minute=0, second=0)
        end_date = last_day_of_month(last_month_dt)

    elif range_str == 'year_to_date':
        start_date = today_start_dt.replace(day=1, month=1, hour=0, minute=0, second=0)
        end_date = today_end_dt

    elif range_str == 'last_year':
        start_date = today_start_dt.replace(year=(today_start_dt.year - 1), day=1,  month=1,  hour=0,  minute=0,  second=0)
        end_date = today_start_dt.replace(year=(today_start_dt.year - 1), day=31, month=12, hour=11, minute=59, second=59)

    return start_date, end_date


class MSPConvertNodeCommand(EnterpriseCommand):
    def get_parser(self):
        return msp_convert_node_parser

    def execute(self, params, **kwargs):
        msp_node_id = None
        node_name = kwargs.get('node')
        if node_name:
            nodes = list(self.resolve_nodes(params, node_name))
            if len(nodes) == 0:
                logging.warning('Node \"%s\" is not found', node_name)
                return
            if len(nodes) > 1:
                logging.warning('More than one nodes \"%s\" are found', node_name)
                return
            msp_node_id = nodes[0]['node_id']
        if msp_node_id is None:
            raise CommandError('msp-convert-node', 'node parameter is required')
        root_nodes = list(self.get_user_root_nodes(params))
        if msp_node_id in root_nodes:
            raise CommandError('msp-convert-node', 'root node cannot be converted')

        node_lookup = {x['node_id']: x for x in params.enterprise.get('nodes', [])}
        role_lookup = {x['role_id']: x for x in params.enterprise.get('roles', [])}
        team_lookup = {x['team_uid']: x for x in params.enterprise.get('teams', [])}
        user_lookup = {x['enterprise_user_id']: x for x in params.enterprise.get('users', [])}

        node_tree = {}    # type: Dict[int, Set[int]]
        for node in node_lookup.values():
            node_id = node['node_id']
            parent_id = node.get('parent_id')
            if isinstance(parent_id, int) and isinstance(node_id, int):
                if parent_id not in node_tree:
                    node_tree[parent_id] = set()
                node_tree[parent_id].add(node_id)
        all_subnodes = [msp_node_id]   # type: List[int]
        pos = 0
        while pos < len(all_subnodes):
            node_id = all_subnodes[pos]
            pos += 1
            if node_id in node_tree:
                all_subnodes.extend(node_tree[node_id])
        nodes_to_move = set(all_subnodes)
        roles_to_move = {x['role_id'] for x in role_lookup.values() if x['node_id'] in nodes_to_move}
        teams_to_move = {x['team_uid'] for x in team_lookup.values() if x['node_id'] in nodes_to_move}
        users_to_move = {x['enterprise_user_id'] for x in user_lookup.values() if x['node_id'] in nodes_to_move}

        errors = []
        for bridge in params.enterprise.get('bridges', []):
            node_id = bridge.get('node_id', 0)
            if node_id in nodes_to_move:
                errors.append(f'Remove bridge provisioning before conversion from node {self.get_node_path(params, node_id)}')
        for scim in params.enterprise.get('scims', []):
            node_id = scim.get('node_id', 0)
            if node_id in nodes_to_move:
                errors.append(f'Remove SCIM provisioning before conversion from node {self.get_node_path(params, node_id)}')
        for sso in params.enterprise.get('sso_services', []):
            node_id = sso.get('node_id', 0)
            if node_id in nodes_to_move:
                errors.append(f'Remove SSO provisioning before conversion from node {self.get_node_path(params, node_id)}')
        for email in params.enterprise.get('email_provision', []):
            node_id = email.get('node_id', 0)
            if node_id in nodes_to_move:
                errors.append(f'Remove email provisioning before conversion from node {self.get_node_path(params, node_id)}')
        for mc in params.enterprise.get('managed_companies', []):
            node_id = mc.get('node_id', 0)
            if node_id in nodes_to_move:
                errors.append(f'Remove managed company before conversion from node {self.get_node_path(params, node_id)}')
        for qt in params.enterprise.get('queued_teams', []):
            node_id = qt['node_id']
            if node_id in nodes_to_move:
                errors.append(f'Remove queued team {qt["name"]} before conversion from node {self.get_node_path(params, node_id)}')

        for user in (y for x, y in user_lookup.items() if x in users_to_move):
            if user['status'] == 'invited':
                errors.append(f'Pending user {user["username"]} must be removed')

        for ru in params.enterprise.get('role_users', []):
            user_id = ru['enterprise_user_id']
            role_id = ru['role_id']
            move_user = user_id in users_to_move
            move_role = role_id in roles_to_move
            if move_role != move_user:
                user = user_lookup.get(user_id)
                username = (user.get('username') if user else '') or str(user_id)
                role = role_lookup.get(role_id)
                rolename = (role['data'].get('displayname') if role else '') or str(role_id)
                errors.append(f'Conflicting role membership: User: {username}, Role: {rolename}')

        for rt in params.enterprise.get('role_teams', []):
            team_uid = rt['team_uid']
            role_id = rt['role_id']
            move_team = team_uid in teams_to_move
            move_role = role_id in roles_to_move
            if move_role != move_team:
                team = team_lookup.get(team_uid)
                teamname = team.get('name', team_uid)
                role = role_lookup.get(role_id)
                rolename = (role['data'].get('displayname') if role else '') or str(role_id)
                errors.append(f'Conflicting role membership: Team: {teamname}, Role: {rolename}')

        for tu in params.enterprise.get('team_users', []):
            user_id = tu['enterprise_user_id']
            team_uid = tu['team_uid']
            move_user = user_id in users_to_move
            move_team = team_uid in teams_to_move
            if move_team != move_user:
                user = user_lookup.get(user_id)
                username = (user['username'] if user else '') or str(user_id)
                team = team_lookup.get(team_uid)
                teamname = (team['name'] if team else '') or team_uid
                errors.append(f'Conflicting team membership: User: {username}, Team: {teamname}')

        for mn in params.enterprise.get('managed_nodes', []):
            role_id = mn['role_id']
            node_id = mn['managed_node_id']
            move_role = role_id in roles_to_move
            move_node = node_id in nodes_to_move
            if move_role != move_node:
                role = role_lookup.get(role_id)
                rolename = (role['data'].get('displayname') if role else '') or str(role_id)
                nodename = self.get_node_path(params, node_id)
                errors.append(f'Conflicting admin role management: Node: {nodename}, Role: {rolename}')

        if len(errors) > 0:
            print('\n'.join(errors))
            return

        seats = kwargs.get('seats') or 0
        if seats < len(users_to_move):
            seats = len(users_to_move)
        if seats == 0:
            seats = 1
        plan = kwargs.get('plan') or 'business'

        msp_node = node_lookup[msp_node_id]
        msp_node_name = msp_node['data'].get('displayname')
        mc = None
        if msp_node_name:
            mc = next(
                (x for x in params.enterprise.get('managed_companies', []) if x['mc_enterprise_name'] == msp_node_name),
                None)
        tree_key = params.enterprise['unencrypted_tree_key']
        if not mc:
            mc_tree_key = utils.generate_aes_key()
            rq = {
                'command': 'enterprise_registration_by_msp',
                'node_id': root_nodes[0],
                'seats': seats,
                'product_id': plan,
                'enterprise_name': msp_node_name,
                'encrypted_tree_key': utils.base64_url_encode(
                    crypto.encrypt_aes_v2(mc_tree_key, tree_key)),
                'role_data': utils.base64_url_encode(
                    crypto.encrypt_aes_v1(json.dumps({'displayname': 'Keeper Administrator'}).encode(), mc_tree_key)),
                'root_node': utils.base64_url_encode(
                    crypto.encrypt_aes_v1(json.dumps({'displayname': 'root'}).encode(), mc_tree_key))
            }
            rs = api.communicate(params, rq)
            mc_id = rs['enterprise_id']
        else:
            mc_id = mc['mc_enterprise_id']

        mc_params = api.login_and_get_mc_params_login_v3(params, mc_id)
        mc_tree_key = mc_params.enterprise['unencrypted_tree_key']

        mc_rq = enterprise_pb2.NodeToManagedCompanyRequest()
        mc_rq.companyId = mc_id
        for node_id in all_subnodes:
            node = node_lookup[node_id]
            red = enterprise_pb2.ReEncryptedData()
            red.id = node_id
            data = json.dumps(node['data']).encode()
            red.data = utils.base64_url_encode(crypto.encrypt_aes_v1(data, mc_tree_key))
            mc_rq.nodes.append(red)

        for role_id in roles_to_move:
            role = role_lookup[role_id]
            red = enterprise_pb2.ReEncryptedData()
            red.id = role_id
            data = json.dumps(role['data']).encode()
            red.data = utils.base64_url_encode(crypto.encrypt_aes_v1(data, mc_tree_key))
            mc_rq.roles.append(red)

        for user_id in users_to_move:
            user = user_lookup[user_id]
            red = enterprise_pb2.ReEncryptedData()
            red.id = user_id
            if user.get('key_type') == 'no_key':
                displayname = user['data'].get('displayname', '?')
                red.data = displayname
            else:
                data = json.dumps(user['data']).encode()
                red.data = utils.base64_url_encode(crypto.encrypt_aes_v1(data, mc_tree_key))
            mc_rq.users.append(red)

        for mn in params.enterprise.get('managed_nodes', []):
            role_id = mn['role_id']
            if role_id in roles_to_move:
                role_key2 = next((x for x in params.enterprise.get('role_keys2', [])), None)
                if role_key2:
                    role_key = utils.base64_url_decode(role_key2['role_key'])
                    role_key = crypto.decrypt_aes_v2(role_key, tree_key)
                    rerk = enterprise_pb2.ReEncryptedRoleKey()
                    rerk.role_id = role_id
                    rerk.encryptedRoleKey = crypto.encrypt_aes_v2(role_key, mc_tree_key)
                    mc_rq.roleKeys.append(rerk)

        for team_uid, team in team_lookup.items():
            if team_uid in teams_to_move:
                etkr = enterprise_pb2.EncryptedTeamKeyRequest()
                etkr.teamUid = utils.base64_url_decode(team_uid)
                encrypted_team_key = team.get('encrypted_team_key')
                if encrypted_team_key:
                    team_key = utils.base64_url_decode(encrypted_team_key)
                    team_key = crypto.decrypt_aes_v2(team_key, tree_key)
                    etkr.encryptedTeamKey = crypto.encrypt_aes_v2(team_key, mc_tree_key)
                else:
                    etkr.force = True
                mc_rq.teamKeys.append(etkr)

        dk_rq = APIRequest_pb2.UserDataKeyByNodeRequest()
        dk_rq.nodeIds.extend(nodes_to_move)
        dk_rs = api.communicate_rest(params, dk_rq, 'enterprise/get_enterprise_user_data_key_by_node',
                                     rs_type=enterprise_pb2.EnterpriseUserDataKeysByNodeResponse)

        if len(dk_rs.keys) > 0:
            keys = params.enterprise['keys']
            mc_keys = mc_params.enterprise['keys']
            if 'ecc_encrypted_private_key' in keys and 'ecc_public_key' in mc_keys:
                encrypted_ec_private_key = utils.base64_url_decode(keys['ecc_encrypted_private_key'])
                ec_private_key = crypto.decrypt_aes_v2(encrypted_ec_private_key, tree_key)
                private_key = crypto.load_ec_private_key(ec_private_key)
                mc_public_key = crypto.load_ec_public_key(utils.base64_url_decode(mc_keys['ecc_public_key']))

                for dk_node in dk_rs.keys:
                    for dk in dk_node.keys:
                        if dk.keyTypeId == 4:
                            enterprise_user_id = dk.enterpriseUserId
                            if enterprise_user_id in users_to_move:
                                encrypted_key = crypto.decrypt_ec(dk.userEncryptedDataKey, private_key)
                                encrypted_key = crypto.encrypt_ec(encrypted_key, mc_public_key)
                                re_dk = enterprise_pb2.ReEncryptedUserDataKey()
                                re_dk.enterpriseUserId = enterprise_user_id
                                re_dk.userEncryptedDataKey = encrypted_key
                                mc_rq.usersDataKeys.append(re_dk)

        api.communicate_rest(params, mc_rq, 'enterprise/node_to_managed_company')
        logging.info(f'Node \"{msp_node_name}\" was converted to Managed Company')
        api.query_enterprise(params)


class MSPCopyRoleCommand(EnterpriseCommand):
    def get_parser(self):
        return msp_copy_role_parser

    def execute(self, params, **kwargs):
        src_roles = {}
        roles = kwargs.get('role')
        if not roles:
            raise CommandError('msp-copy-role', f'Source role parameter is required')

        if not isinstance(roles, list):
            roles = [roles]

        for role_name in roles:
            if not isinstance(role_name, str):
                role_name = str(role_name)
            matched_roles = list(MSPCopyRoleCommand.find_roles(params, role_name))
            if len(matched_roles) == 1:
                role = matched_roles[0]
                src_roles[role['role_id']] = role
            elif len(matched_roles) > 1:
                raise Exception(f'There are more than one roles with name \"{role_name}\". Use Role ID')
            else:
                raise Exception(f'Role \"{role_name}\" not found')

        managed_companies = params.enterprise.get('managed_companies', [])
        mcs = {}
        for mc_name in kwargs.get('mc') or []:
            mc = get_mc_by_name_or_id(managed_companies, mc_name)
            if not mc:
                raise CommandError('msp-copy-role', f'Managed Company \"{mc_name}\" not found')
            mcs[mc['mc_enterprise_id']] = mc

        for mc in mcs.values():
            mc_id = mc['mc_enterprise_id']
            mc_params = api.login_and_get_mc_params_login_v3(params, mc_id)
            node_id = next((x['node_id'] for x in mc_params.enterprise.get('nodes', []) if not x.get('parent_id')), None)
            mc_rqs = []
            for role in src_roles.values():
                src_role_id = role['role_id']
                role_name = role['data'].get('displayname') or ''
                if not role_name:
                    continue
                dst_roles = list(MSPCopyRoleCommand.find_roles(mc_params, role_name))
                if len(dst_roles) > 1:
                    logging.warning('MC # %d: There are more than one roles with name \"%s\". Skipping', mc_id, role_name)
                    continue

                if len(dst_roles) == 1:
                    dst_role_id = dst_roles[0]['role_id']
                else:
                    dst_role_id = self.get_enterprise_id(mc_params)
                    dt = json.dumps({ "displayname": role_name })
                    mc_rqs.append({
                        "command": 'role_add',
                        "role_id": dst_role_id,
                        "node_id": node_id,
                        "encrypted_data": utils.base64_url_encode(
                            crypto.encrypt_aes_v1(dt.encode('utf-8'), mc_params.enterprise['unencrypted_tree_key'])),
                        "visible_below": role.get('visible_below', True),
                        "new_user_inherit": role.get('new_user_inherit',  False)
                    })
                enf = next((x['enforcements'] for x in params.enterprise.get('role_enforcements') or [] if x.get('role_id') == src_role_id), None)
                src_enforcements = enf.copy() if isinstance(enf, dict) else {}
                enf = next((x['enforcements'] for x in mc_params.enterprise.get('role_enforcements') or [] if x.get('role_id') == dst_role_id), None)
                dst_enforcements = enf.copy() if isinstance(enf, dict) else {}
                for enforcement in src_enforcements:
                    src_value = src_enforcements[enforcement]
                    if enforcement in dst_enforcements:
                        dst_value = dst_enforcements[enforcement]
                        dst_enforcements.pop(enforcement)
                        if src_value == dst_value:
                            continue
                        command = 'role_enforcement_update'
                    else:
                        command = 'role_enforcement_add'
                    if command:
                        rq = {
                            'command': command,
                            'role_id': dst_role_id,
                            'enforcement': enforcement,
                        }
                        try:
                            value = MSPCopyRoleCommand.get_enforcement_value(enforcement, src_value)
                            if value is not None:
                                if not isinstance(value, bool):
                                    rq['value'] = value
                                mc_rqs.append(rq)
                        except Exception as e:
                            logging.warning('Role %s: Enforcement %s: %s', role_name, enforcement, e)
                for enforcement in dst_enforcements:
                    rq = {
                        'command': 'role_enforcement_remove',
                        'role_id': dst_role_id,
                        'enforcement': enforcement,
                    }
                    mc_rqs.append(rq)
            if mc_rqs:
                api.execute_batch(mc_params, mc_rqs)
            logging.info('MC %s: Roles are in sync', mc_id)

    @staticmethod
    def get_enforcement_value(name, value):    # type: (str, str) -> Any
        name = name.lower()
        if name in constants.ENFORCEMENTS:
            enforcement_type = constants.ENFORCEMENTS[name]
            if enforcement_type == 'long':
                try:
                    return int(value)
                except Exception as e:
                    raise Exception(f'Enforcement {name}: invalid integer value: {value}')
            if enforcement_type == 'boolean':
                return value == 'true'
            if enforcement_type == 'account_share':  # not supported
                return
            if enforcement_type in ('record_types', 'json', 'jsonarray'):
                return json.loads(value)

            return value  # 'ip_whitelist', 'string', 'two_factor_duration', 'ternary_*'

    @staticmethod
    def find_roles(params, name):   # type: (KeeperParams, str) -> Iterable[Dict]
        if isinstance(params.enterprise, dict):
            if name.isdigit():
                role_id = int(name)
                for role in params.enterprise.get('roles') or []:
                    if role_id == role.get('role_id'):
                        yield role
                        return

            for role in params.enterprise.get('roles') or []:
                role_name = role['data'].get('displayname') or ''
                if role_name.casefold() == name.casefold():
                    yield role



from typing import Optional, BinaryIO, Tuple

from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKey, EllipticCurvePublicKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey, RSAPrivateKey


def get_random_bytes(length: int) -> bytes: ...

def pad_data(data: bytes) -> bytes: ...
def unpad_data(data: bytes) -> bytes: ...

def generate_rsa_key() -> Tuple[RSAPrivateKey, RSAPublicKey]: ...
def load_rsa_private_key(der_private_key: bytes, password: Optional[str] = ...) -> RSAPrivateKey: ...
def load_rsa_public_key(der_public_key: bytes) -> RSAPublicKey: ...
def unload_rsa_private_key(private_key: RSAPrivateKey) -> bytes: ...
def unload_rsa_public_key(public_key: RSAPublicKey) -> bytes: ...

def generate_ec_key() -> (EllipticCurvePrivateKey, EllipticCurvePublicKey): ...
def load_ec_private_key(private_key: bytes) -> EllipticCurvePrivateKey: ...
def load_ec_public_key(public_key: bytes) -> EllipticCurvePublicKey: ...
def unload_ec_private_key(private_key: EllipticCurvePrivateKey) -> bytes: ...
def unload_ec_public_key(public_key: EllipticCurvePublicKey) -> bytes: ...

def encrypt_aes_v1(data: bytes, key: bytes, iv: Optional[bytes] = ..., use_padding: bool = ...) -> bytes: ...
def decrypt_aes_v1(data: bytes, key: bytes, use_padding: bool = ...) -> bytes: ...

def encrypt_aes_v2(data: bytes, key: bytes, nonce: bytes = ...) -> bytes: ...
def decrypt_aes_v2(data: bytes, key: bytes) -> bytes: ...

def encrypt_rsa(data: bytes, rsa_key: RSAPublicKey) -> bytes: ...
def decrypt_rsa(data: bytes, rsa_key: RSAPrivateKey) -> bytes: ...

def encrypt_ec(data: bytes, ec_public_key: EllipticCurvePublicKey) -> bytes: ...
def decrypt_ec(data: bytes, ec_private_key: EllipticCurvePrivateKey) -> bytes: ...

def derive_key_v1(password: str, salt: bytes, iterations: int) -> bytes: ...
def derive_keyhash_v1(password: str, salt: bytes, iterations: int) -> bytes: ...
def derive_keyhash_v2(domain: str, password: str, salt: bytes, iterations: int) -> bytes: ...

def hmac_sha512(key: bytes, data: bytes) -> bytes: ...

def generate_hkdf_key(info: str, phrase: str) -> bytes: ...

class StreamCrypter:
    key: bytes
    is_gcm: bool
    @property
    def bytes_read(self) -> int: ...
    def set_stream(self, stream: BinaryIO, for_encrypt: bool) -> BinaryIO: ...





from unittest import TestCase

import pytest
from keepercommander.commands import scim

@pytest.mark.skip
class TestScimCommands(TestCase):
    @staticmethod
    def add_scim_team():
        scim_url = 'SCIM URL'
        token = 'SCIM TOKEN'
        payload = {
            "schemas": ["urn:ietf:params:scim:schemas:core:2.0:Group"],
            "externalId":"e9e306a331660a",
            "displayName": "Queued Team",
        }
        scim.ScimPushCommand.post_scim_resource(f'{scim_url}/Groups', token, payload)



import ctypes
import os
import shutil
import sys
import subprocess
import platform

is_windows = sys.platform.startswith('win')
is_macos = sys.platform.startswith('darwin')
is_linux = sys.platform.startswith('linux')

if is_windows:
    # Constants
    LOGON_WITH_PROFILE = 0x00000001
    CREATE_UNICODE_ENVIRONMENT = 0x00000400

    class STARTUPINFO(ctypes.Structure):
        _fields_ = [
            ("cb", ctypes.c_ulong),
            ("lpReserved", ctypes.c_wchar_p),
            ("lpDesktop", ctypes.c_wchar_p),
            ("lpTitle", ctypes.c_wchar_p),
            ("dwX", ctypes.c_ulong),
            ("dwY", ctypes.c_ulong),
            ("dwXSize", ctypes.c_ulong),
            ("dwYSize", ctypes.c_ulong),
            ("dwXCountChars", ctypes.c_ulong),
            ("dwYCountChars", ctypes.c_ulong),
            ("dwFillAttribute", ctypes.c_ulong),
            ("dwFlags", ctypes.c_ulong),
            ("wShowWindow", ctypes.c_ushort),
            ("cbReserved2", ctypes.c_ushort),
            ("lpReserved2", ctypes.c_void_p),
            ("hStdInput", ctypes.c_void_p),
            ("hStdOutput", ctypes.c_void_p),
            ("hStdError", ctypes.c_void_p),
        ]

    class PROCESS_INFORMATION(ctypes.Structure):
        _fields_ = [
            ("hProcess", ctypes.c_void_p),
            ("hThread", ctypes.c_void_p),
            ("dwProcessId", ctypes.c_ulong),
            ("dwThreadId", ctypes.c_ulong),
        ]

    # Load function
    CreateProcessWithLogonW = ctypes.windll.advapi32.CreateProcessWithLogonW
    CreateProcessWithLogonW.argtypes = [
        ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_wchar_p,
        ctypes.c_ulong, ctypes.c_wchar_p, ctypes.c_wchar_p,
        ctypes.c_ulong, ctypes.c_void_p, ctypes.c_wchar_p,
        ctypes.POINTER(STARTUPINFO), ctypes.POINTER(PROCESS_INFORMATION)
    ]
    CreateProcessWithLogonW.restype = ctypes.c_bool

    FORMAT_MESSAGE_ALLOCATE_BUFFER = 0x00000100
    FORMAT_MESSAGE_FROM_SYSTEM = 0x00001000
    FORMAT_MESSAGE_IGNORE_INSERTS = 0x00000200

    # Get the error message
    flags = FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS

    def get_error_message(error_code):    # type: (int) -> str
        message_buffer = ctypes.create_unicode_buffer(0)
        length = ctypes.windll.kernel32.FormatMessageW(
            flags,
            None,  # Message source
            error_code,  # Message identifier
            0,  # Language ID
            ctypes.byref(message_buffer),  # Buffer
            0,  # Size
            None  # Arguments
        )
        if length:
            return message_buffer.value.strip()
        return 'Unknown'


def run_as(username, password, application):    # type: (str, str, str) -> None
    if is_windows:
        if '@' in username:
            domain = ''
        else:
            domain, sep, username = username.rpartition('\\')
            if not domain:
                domain = '.'

        application_path = shutil.which(application)
        if not application_path:
            raise Exception(f'Application "{application}" not found. Please use full application path')

        si = STARTUPINFO()
        si.cb = ctypes.sizeof(STARTUPINFO)
        pi = PROCESS_INFORMATION()

        # Launch process
        success = CreateProcessWithLogonW(
            username,
            domain,
            password,
            LOGON_WITH_PROFILE,
            None,
            f'"{application_path}"',
            CREATE_UNICODE_ENVIRONMENT,
            None,
            None,
            ctypes.byref(si),
            ctypes.byref(pi)
        )

        if not success:
            error_code = ctypes.GetLastError()
            raise OSError(f'Failed to launch "{application}" as {username}. Windows Error: {error_code}')
    elif is_macos or is_linux:
        # Find the full path to the application
        application_path = shutil.which(application)
        if not application_path:
            raise Exception(f'Application "{application}" not found. Please use full application path')

        # Check if user exists
        try:
            subprocess.run(['id', '-u', username], check=True, capture_output=True)
        except subprocess.CalledProcessError:
            raise OSError(f'User "{username}" does not exist')

        # Use sudo if we're not root, otherwise use su
        if os.geteuid() != 0:
            cmd = ['sudo', '-u', username, application_path]
            proc = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            proc.communicate(input=f'{password}\n'.encode())
        else:
            cmd = ['su', '-', username, '-c', application_path]
            proc = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            proc.communicate(input=f'{password}\n'.encode())

        if proc.returncode != 0:
            raise OSError(f'Failed to launch "{application}" as {username}. Error code: {proc.returncode}')
    else:
        raise OSError('Unsupported operating system')



import APIRequest_pb2 as _APIRequest_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class DeviceActionType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    DA_INVALID: _ClassVar[DeviceActionType]
    DA_LOGOUT: _ClassVar[DeviceActionType]
    DA_REMOVE: _ClassVar[DeviceActionType]
    DA_LOCK: _ClassVar[DeviceActionType]
    DA_UNLOCK: _ClassVar[DeviceActionType]
    DA_DEVICE_ACCOUNT_LOCK: _ClassVar[DeviceActionType]
    DA_DEVICE_ACCOUNT_UNLOCK: _ClassVar[DeviceActionType]
    DA_LINK: _ClassVar[DeviceActionType]
    DA_UNLINK: _ClassVar[DeviceActionType]

class DeviceActionStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    INVALID: _ClassVar[DeviceActionStatus]
    SUCCESS: _ClassVar[DeviceActionStatus]
    NOT_ALLOWED: _ClassVar[DeviceActionStatus]

class ClientType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    NONE: _ClassVar[ClientType]
    ANDROID: _ClassVar[ClientType]
    BLACKBERRY: _ClassVar[ClientType]
    BLACKBERRY10: _ClassVar[ClientType]
    DESKTOP: _ClassVar[ClientType]
    IOS: _ClassVar[ClientType]
    MAC_APP: _ClassVar[ClientType]
    WEB_APP: _ClassVar[ClientType]
    WINDOWS_PHONE: _ClassVar[ClientType]
    SURFACE: _ClassVar[ClientType]
    WIN8: _ClassVar[ClientType]
    IE_EXTENSION: _ClassVar[ClientType]
    CHROME_EXTENSION: _ClassVar[ClientType]
    FIREFOX_EXTENSION: _ClassVar[ClientType]
    SAFARI_EXTENSION: _ClassVar[ClientType]
    DOCOMO: _ClassVar[ClientType]
    UNKNOWN: _ClassVar[ClientType]
    SERVER: _ClassVar[ClientType]
    COMMANDER: _ClassVar[ClientType]
    BRIDGE: _ClassVar[ClientType]
    ENTERPRISE_MANAGEMENT_CONSOLE: _ClassVar[ClientType]
    EDGE_EXTENSION: _ClassVar[ClientType]
    SUPPORT_TOOL: _ClassVar[ClientType]
    SSO_CONNECT: _ClassVar[ClientType]
    DESKTOP_ELECTRON: _ClassVar[ClientType]
    PASSWORD_IMPORTER: _ClassVar[ClientType]
    CHAT: _ClassVar[ClientType]
    MAC_APP_ELECTRON: _ClassVar[ClientType]
    CHAT_IOS: _ClassVar[ClientType]
    CHAT_ANDROID: _ClassVar[ClientType]
    CHAT_WINDOWS: _ClassVar[ClientType]
    CHAT_MAC: _ClassVar[ClientType]
    SCIM: _ClassVar[ClientType]
    LAMBDA: _ClassVar[ClientType]
    CONNECTWISE_CONTROL_HELPER: _ClassVar[ClientType]
    ENTERPRISE_CLIENT_TOOL: _ClassVar[ClientType]
    SECRETS_MANAGER_JS: _ClassVar[ClientType]
    SECRETS_MANAGER_PYTHON: _ClassVar[ClientType]
    SECRETS_MANAGER_GO: _ClassVar[ClientType]
    SECRETS_MANAGER_JAVA: _ClassVar[ClientType]
    SECRETS_MANAGER_NET: _ClassVar[ClientType]

class ClientTypeCategory(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    CAT_NONE: _ClassVar[ClientTypeCategory]
    CAT_ADMIN: _ClassVar[ClientTypeCategory]
    CAT_DESKTOP: _ClassVar[ClientTypeCategory]
    CAT_EXTENSION: _ClassVar[ClientTypeCategory]
    CAT_MOBILE: _ClassVar[ClientTypeCategory]
    CAT_OTHER: _ClassVar[ClientTypeCategory]
    CAT_WEB_VAULT: _ClassVar[ClientTypeCategory]
    CAT_CHAT_DESKTOP: _ClassVar[ClientTypeCategory]
    CAT_CHAT_MOBILE: _ClassVar[ClientTypeCategory]
    CAT_SECRETS_MANAGER: _ClassVar[ClientTypeCategory]
DA_INVALID: DeviceActionType
DA_LOGOUT: DeviceActionType
DA_REMOVE: DeviceActionType
DA_LOCK: DeviceActionType
DA_UNLOCK: DeviceActionType
DA_DEVICE_ACCOUNT_LOCK: DeviceActionType
DA_DEVICE_ACCOUNT_UNLOCK: DeviceActionType
DA_LINK: DeviceActionType
DA_UNLINK: DeviceActionType
INVALID: DeviceActionStatus
SUCCESS: DeviceActionStatus
NOT_ALLOWED: DeviceActionStatus
NONE: ClientType
ANDROID: ClientType
BLACKBERRY: ClientType
BLACKBERRY10: ClientType
DESKTOP: ClientType
IOS: ClientType
MAC_APP: ClientType
WEB_APP: ClientType
WINDOWS_PHONE: ClientType
SURFACE: ClientType
WIN8: ClientType
IE_EXTENSION: ClientType
CHROME_EXTENSION: ClientType
FIREFOX_EXTENSION: ClientType
SAFARI_EXTENSION: ClientType
DOCOMO: ClientType
UNKNOWN: ClientType
SERVER: ClientType
COMMANDER: ClientType
BRIDGE: ClientType
ENTERPRISE_MANAGEMENT_CONSOLE: ClientType
EDGE_EXTENSION: ClientType
SUPPORT_TOOL: ClientType
SSO_CONNECT: ClientType
DESKTOP_ELECTRON: ClientType
PASSWORD_IMPORTER: ClientType
CHAT: ClientType
MAC_APP_ELECTRON: ClientType
CHAT_IOS: ClientType
CHAT_ANDROID: ClientType
CHAT_WINDOWS: ClientType
CHAT_MAC: ClientType
SCIM: ClientType
LAMBDA: ClientType
CONNECTWISE_CONTROL_HELPER: ClientType
ENTERPRISE_CLIENT_TOOL: ClientType
SECRETS_MANAGER_JS: ClientType
SECRETS_MANAGER_PYTHON: ClientType
SECRETS_MANAGER_GO: ClientType
SECRETS_MANAGER_JAVA: ClientType
SECRETS_MANAGER_NET: ClientType
CAT_NONE: ClientTypeCategory
CAT_ADMIN: ClientTypeCategory
CAT_DESKTOP: ClientTypeCategory
CAT_EXTENSION: ClientTypeCategory
CAT_MOBILE: ClientTypeCategory
CAT_OTHER: ClientTypeCategory
CAT_WEB_VAULT: ClientTypeCategory
CAT_CHAT_DESKTOP: ClientTypeCategory
CAT_CHAT_MOBILE: ClientTypeCategory
CAT_SECRETS_MANAGER: ClientTypeCategory

class Device(_message.Message):
    __slots__ = ("encryptedDeviceToken", "deviceName", "devicePlatform", "deviceStatus", "loginState", "clientVersion", "clientType", "clientTypeCategory", "clientFormFactor", "lastModifiedTime")
    ENCRYPTEDDEVICETOKEN_FIELD_NUMBER: _ClassVar[int]
    DEVICENAME_FIELD_NUMBER: _ClassVar[int]
    DEVICEPLATFORM_FIELD_NUMBER: _ClassVar[int]
    DEVICESTATUS_FIELD_NUMBER: _ClassVar[int]
    LOGINSTATE_FIELD_NUMBER: _ClassVar[int]
    CLIENTVERSION_FIELD_NUMBER: _ClassVar[int]
    CLIENTTYPE_FIELD_NUMBER: _ClassVar[int]
    CLIENTTYPECATEGORY_FIELD_NUMBER: _ClassVar[int]
    CLIENTFORMFACTOR_FIELD_NUMBER: _ClassVar[int]
    LASTMODIFIEDTIME_FIELD_NUMBER: _ClassVar[int]
    encryptedDeviceToken: bytes
    deviceName: str
    devicePlatform: str
    deviceStatus: _APIRequest_pb2.DeviceStatus
    loginState: _APIRequest_pb2.LoginState
    clientVersion: str
    clientType: ClientType
    clientTypeCategory: ClientTypeCategory
    clientFormFactor: _APIRequest_pb2.ClientFormFactor
    lastModifiedTime: int
    def __init__(self, encryptedDeviceToken: _Optional[bytes] = ..., deviceName: _Optional[str] = ..., devicePlatform: _Optional[str] = ..., deviceStatus: _Optional[_Union[_APIRequest_pb2.DeviceStatus, str]] = ..., loginState: _Optional[_Union[_APIRequest_pb2.LoginState, str]] = ..., clientVersion: _Optional[str] = ..., clientType: _Optional[_Union[ClientType, str]] = ..., clientTypeCategory: _Optional[_Union[ClientTypeCategory, str]] = ..., clientFormFactor: _Optional[_Union[_APIRequest_pb2.ClientFormFactor, str]] = ..., lastModifiedTime: _Optional[int] = ...) -> None: ...

class DeviceGroup(_message.Message):
    __slots__ = ("devices",)
    DEVICES_FIELD_NUMBER: _ClassVar[int]
    devices: _containers.RepeatedCompositeFieldContainer[Device]
    def __init__(self, devices: _Optional[_Iterable[_Union[Device, _Mapping]]] = ...) -> None: ...

class DeviceUserResponse(_message.Message):
    __slots__ = ("deviceGroups",)
    DEVICEGROUPS_FIELD_NUMBER: _ClassVar[int]
    deviceGroups: _containers.RepeatedCompositeFieldContainer[DeviceGroup]
    def __init__(self, deviceGroups: _Optional[_Iterable[_Union[DeviceGroup, _Mapping]]] = ...) -> None: ...

class DeviceActionRequest(_message.Message):
    __slots__ = ("deviceAction",)
    DEVICEACTION_FIELD_NUMBER: _ClassVar[int]
    deviceAction: _containers.RepeatedCompositeFieldContainer[DeviceAction]
    def __init__(self, deviceAction: _Optional[_Iterable[_Union[DeviceAction, _Mapping]]] = ...) -> None: ...

class DeviceAction(_message.Message):
    __slots__ = ("deviceActionType", "encryptedDeviceToken")
    DEVICEACTIONTYPE_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDDEVICETOKEN_FIELD_NUMBER: _ClassVar[int]
    deviceActionType: DeviceActionType
    encryptedDeviceToken: _containers.RepeatedScalarFieldContainer[bytes]
    def __init__(self, deviceActionType: _Optional[_Union[DeviceActionType, str]] = ..., encryptedDeviceToken: _Optional[_Iterable[bytes]] = ...) -> None: ...

class DeviceActionResponse(_message.Message):
    __slots__ = ("deviceActionResult", "deviceGroups")
    DEVICEACTIONRESULT_FIELD_NUMBER: _ClassVar[int]
    DEVICEGROUPS_FIELD_NUMBER: _ClassVar[int]
    deviceActionResult: _containers.RepeatedCompositeFieldContainer[DeviceActionResult]
    deviceGroups: _containers.RepeatedCompositeFieldContainer[DeviceGroup]
    def __init__(self, deviceActionResult: _Optional[_Iterable[_Union[DeviceActionResult, _Mapping]]] = ..., deviceGroups: _Optional[_Iterable[_Union[DeviceGroup, _Mapping]]] = ...) -> None: ...

class DeviceActionResult(_message.Message):
    __slots__ = ("deviceActionType", "encryptedDeviceToken", "deviceActionStatus")
    DEVICEACTIONTYPE_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDDEVICETOKEN_FIELD_NUMBER: _ClassVar[int]
    DEVICEACTIONSTATUS_FIELD_NUMBER: _ClassVar[int]
    deviceActionType: DeviceActionType
    encryptedDeviceToken: _containers.RepeatedScalarFieldContainer[bytes]
    deviceActionStatus: DeviceActionStatus
    def __init__(self, deviceActionType: _Optional[_Union[DeviceActionType, str]] = ..., encryptedDeviceToken: _Optional[_Iterable[bytes]] = ..., deviceActionStatus: _Optional[_Union[DeviceActionStatus, str]] = ...) -> None: ...

class DeviceRenameRequest(_message.Message):
    __slots__ = ("deviceRename",)
    DEVICERENAME_FIELD_NUMBER: _ClassVar[int]
    deviceRename: _containers.RepeatedCompositeFieldContainer[DeviceRename]
    def __init__(self, deviceRename: _Optional[_Iterable[_Union[DeviceRename, _Mapping]]] = ...) -> None: ...

class DeviceRename(_message.Message):
    __slots__ = ("encryptedDeviceToken", "deviceNewName")
    ENCRYPTEDDEVICETOKEN_FIELD_NUMBER: _ClassVar[int]
    DEVICENEWNAME_FIELD_NUMBER: _ClassVar[int]
    encryptedDeviceToken: bytes
    deviceNewName: str
    def __init__(self, encryptedDeviceToken: _Optional[bytes] = ..., deviceNewName: _Optional[str] = ...) -> None: ...

class DeviceRenameResponse(_message.Message):
    __slots__ = ("deviceRenameResult", "deviceGroups")
    DEVICERENAMERESULT_FIELD_NUMBER: _ClassVar[int]
    DEVICEGROUPS_FIELD_NUMBER: _ClassVar[int]
    deviceRenameResult: _containers.RepeatedCompositeFieldContainer[DeviceRenameResult]
    deviceGroups: _containers.RepeatedCompositeFieldContainer[DeviceGroup]
    def __init__(self, deviceRenameResult: _Optional[_Iterable[_Union[DeviceRenameResult, _Mapping]]] = ..., deviceGroups: _Optional[_Iterable[_Union[DeviceGroup, _Mapping]]] = ...) -> None: ...

class DeviceRenameResult(_message.Message):
    __slots__ = ("encryptedDeviceToken", "deviceNewName", "deviceActionStatus")
    ENCRYPTEDDEVICETOKEN_FIELD_NUMBER: _ClassVar[int]
    DEVICENEWNAME_FIELD_NUMBER: _ClassVar[int]
    DEVICEACTIONSTATUS_FIELD_NUMBER: _ClassVar[int]
    encryptedDeviceToken: bytes
    deviceNewName: str
    deviceActionStatus: DeviceActionStatus
    def __init__(self, encryptedDeviceToken: _Optional[bytes] = ..., deviceNewName: _Optional[str] = ..., deviceActionStatus: _Optional[_Union[DeviceActionStatus, str]] = ...) -> None: ...

class DeviceAdminRequest(_message.Message):
    __slots__ = ("enterpriseUserIds",)
    ENTERPRISEUSERIDS_FIELD_NUMBER: _ClassVar[int]
    enterpriseUserIds: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, enterpriseUserIds: _Optional[_Iterable[int]] = ...) -> None: ...

class DeviceAdminResponse(_message.Message):
    __slots__ = ("deviceUserList",)
    DEVICEUSERLIST_FIELD_NUMBER: _ClassVar[int]
    deviceUserList: _containers.RepeatedCompositeFieldContainer[DeviceUserList]
    def __init__(self, deviceUserList: _Optional[_Iterable[_Union[DeviceUserList, _Mapping]]] = ...) -> None: ...

class DeviceUserList(_message.Message):
    __slots__ = ("enterpriseUserId", "deviceGroups")
    ENTERPRISEUSERID_FIELD_NUMBER: _ClassVar[int]
    DEVICEGROUPS_FIELD_NUMBER: _ClassVar[int]
    enterpriseUserId: int
    deviceGroups: _containers.RepeatedCompositeFieldContainer[DeviceGroup]
    def __init__(self, enterpriseUserId: _Optional[int] = ..., deviceGroups: _Optional[_Iterable[_Union[DeviceGroup, _Mapping]]] = ...) -> None: ...

class DeviceAdminActionRequest(_message.Message):
    __slots__ = ("deviceAdminAction",)
    DEVICEADMINACTION_FIELD_NUMBER: _ClassVar[int]
    deviceAdminAction: _containers.RepeatedCompositeFieldContainer[DeviceAdminAction]
    def __init__(self, deviceAdminAction: _Optional[_Iterable[_Union[DeviceAdminAction, _Mapping]]] = ...) -> None: ...

class DeviceAdminAction(_message.Message):
    __slots__ = ("deviceActionType", "enterpriseUserId", "encryptedDeviceToken")
    DEVICEACTIONTYPE_FIELD_NUMBER: _ClassVar[int]
    ENTERPRISEUSERID_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDDEVICETOKEN_FIELD_NUMBER: _ClassVar[int]
    deviceActionType: DeviceActionType
    enterpriseUserId: int
    encryptedDeviceToken: _containers.RepeatedScalarFieldContainer[bytes]
    def __init__(self, deviceActionType: _Optional[_Union[DeviceActionType, str]] = ..., enterpriseUserId: _Optional[int] = ..., encryptedDeviceToken: _Optional[_Iterable[bytes]] = ...) -> None: ...

class DeviceAdminActionResponse(_message.Message):
    __slots__ = ("deviceAdminActionResults",)
    DEVICEADMINACTIONRESULTS_FIELD_NUMBER: _ClassVar[int]
    deviceAdminActionResults: _containers.RepeatedCompositeFieldContainer[DeviceAdminActionResult]
    def __init__(self, deviceAdminActionResults: _Optional[_Iterable[_Union[DeviceAdminActionResult, _Mapping]]] = ...) -> None: ...

class DeviceAdminActionResult(_message.Message):
    __slots__ = ("deviceActionType", "enterpriseUserId", "encryptedDeviceToken", "deviceActionStatus")
    DEVICEACTIONTYPE_FIELD_NUMBER: _ClassVar[int]
    ENTERPRISEUSERID_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDDEVICETOKEN_FIELD_NUMBER: _ClassVar[int]
    DEVICEACTIONSTATUS_FIELD_NUMBER: _ClassVar[int]
    deviceActionType: DeviceActionType
    enterpriseUserId: int
    encryptedDeviceToken: _containers.RepeatedScalarFieldContainer[bytes]
    deviceActionStatus: DeviceActionStatus
    def __init__(self, deviceActionType: _Optional[_Union[DeviceActionType, str]] = ..., enterpriseUserId: _Optional[int] = ..., encryptedDeviceToken: _Optional[_Iterable[bytes]] = ..., deviceActionStatus: _Optional[_Union[DeviceActionStatus, str]] = ...) -> None: ...



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class SecurityBenchmark(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    SB_INVALID: _ClassVar[SecurityBenchmark]
    SB_DEPLOY_ACROSS_ENTIRE_ORGANIZATION: _ClassVar[SecurityBenchmark]
    SB_PREVENT_INSTALLATION_OF_UNTRUSTED_EXTENSIONS: _ClassVar[SecurityBenchmark]
    SB_ENABLE_ACCOUNT_TRANSFER_POLICY: _ClassVar[SecurityBenchmark]
    SB_REDUCE_ADMINISTRATOR_PRIVILEGE: _ClassVar[SecurityBenchmark]
    SB_ENSURE_OUTSIDE_SSO_ADMINISTRATOR_EXISTS: _ClassVar[SecurityBenchmark]
    SB_LOCK_DOWN_SSO_PROVIDER: _ClassVar[SecurityBenchmark]
    SB_CREATE_AT_LEAST_TWO_KEEPER_ADMINISTRATORS: _ClassVar[SecurityBenchmark]
    SB_ENSURE_TWO_FACTOR_AUTHENTICATION_ADMIN_USERS: _ClassVar[SecurityBenchmark]
    SB_ENFORCE_STRONG_MASTER_PASSWORD: _ClassVar[SecurityBenchmark]
    SB_ENSURE_TWO_FACTOR_AUTHENTICATION_FOR_END_USERS: _ClassVar[SecurityBenchmark]
    SB_CONFIGURE_IP_ALLOWLISTING: _ClassVar[SecurityBenchmark]
    SB_CREATE_ALERTS: _ClassVar[SecurityBenchmark]
    SB_DISABLE_BROWSER_PASSWORD_MANAGERS: _ClassVar[SecurityBenchmark]
    SB_DISABLE_ACCOUNT_RECOVERY: _ClassVar[SecurityBenchmark]

class SecurityBenchmarkStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    INVALID: _ClassVar[SecurityBenchmarkStatus]
    RESOLVED: _ClassVar[SecurityBenchmarkStatus]
    IGNORED: _ClassVar[SecurityBenchmarkStatus]
    UNRESOLVED: _ClassVar[SecurityBenchmarkStatus]
SB_INVALID: SecurityBenchmark
SB_DEPLOY_ACROSS_ENTIRE_ORGANIZATION: SecurityBenchmark
SB_PREVENT_INSTALLATION_OF_UNTRUSTED_EXTENSIONS: SecurityBenchmark
SB_ENABLE_ACCOUNT_TRANSFER_POLICY: SecurityBenchmark
SB_REDUCE_ADMINISTRATOR_PRIVILEGE: SecurityBenchmark
SB_ENSURE_OUTSIDE_SSO_ADMINISTRATOR_EXISTS: SecurityBenchmark
SB_LOCK_DOWN_SSO_PROVIDER: SecurityBenchmark
SB_CREATE_AT_LEAST_TWO_KEEPER_ADMINISTRATORS: SecurityBenchmark
SB_ENSURE_TWO_FACTOR_AUTHENTICATION_ADMIN_USERS: SecurityBenchmark
SB_ENFORCE_STRONG_MASTER_PASSWORD: SecurityBenchmark
SB_ENSURE_TWO_FACTOR_AUTHENTICATION_FOR_END_USERS: SecurityBenchmark
SB_CONFIGURE_IP_ALLOWLISTING: SecurityBenchmark
SB_CREATE_ALERTS: SecurityBenchmark
SB_DISABLE_BROWSER_PASSWORD_MANAGERS: SecurityBenchmark
SB_DISABLE_ACCOUNT_RECOVERY: SecurityBenchmark
INVALID: SecurityBenchmarkStatus
RESOLVED: SecurityBenchmarkStatus
IGNORED: SecurityBenchmarkStatus
UNRESOLVED: SecurityBenchmarkStatus

class EnterpriseStat(_message.Message):
    __slots__ = ["usersLoggedRecent", "usersHasRecords"]
    USERSLOGGEDRECENT_FIELD_NUMBER: _ClassVar[int]
    USERSHASRECORDS_FIELD_NUMBER: _ClassVar[int]
    usersLoggedRecent: int
    usersHasRecords: int
    def __init__(self, usersLoggedRecent: _Optional[int] = ..., usersHasRecords: _Optional[int] = ...) -> None: ...

class EnterpriseStatDetail(_message.Message):
    __slots__ = ["enterpriseUserId", "lastLoggedIn", "hasRecords"]
    ENTERPRISEUSERID_FIELD_NUMBER: _ClassVar[int]
    LASTLOGGEDIN_FIELD_NUMBER: _ClassVar[int]
    HASRECORDS_FIELD_NUMBER: _ClassVar[int]
    enterpriseUserId: int
    lastLoggedIn: int
    hasRecords: bool
    def __init__(self, enterpriseUserId: _Optional[int] = ..., lastLoggedIn: _Optional[int] = ..., hasRecords: bool = ...) -> None: ...

class EnterpriseStatContinuationToken(_message.Message):
    __slots__ = ["lastUpdated", "enterpriseUserId"]
    LASTUPDATED_FIELD_NUMBER: _ClassVar[int]
    ENTERPRISEUSERID_FIELD_NUMBER: _ClassVar[int]
    lastUpdated: int
    enterpriseUserId: int
    def __init__(self, lastUpdated: _Optional[int] = ..., enterpriseUserId: _Optional[int] = ...) -> None: ...

class EnterpriseStatDetailsRequest(_message.Message):
    __slots__ = ["lastUpdated", "continuationToken"]
    LASTUPDATED_FIELD_NUMBER: _ClassVar[int]
    CONTINUATIONTOKEN_FIELD_NUMBER: _ClassVar[int]
    lastUpdated: int
    continuationToken: EnterpriseStatContinuationToken
    def __init__(self, lastUpdated: _Optional[int] = ..., continuationToken: _Optional[_Union[EnterpriseStatContinuationToken, _Mapping]] = ...) -> None: ...

class EnterpriseStatDetailsResponse(_message.Message):
    __slots__ = ["enterpriseStatDetails", "lastUpdated", "continuationToken", "hasMore"]
    ENTERPRISESTATDETAILS_FIELD_NUMBER: _ClassVar[int]
    LASTUPDATED_FIELD_NUMBER: _ClassVar[int]
    CONTINUATIONTOKEN_FIELD_NUMBER: _ClassVar[int]
    HASMORE_FIELD_NUMBER: _ClassVar[int]
    enterpriseStatDetails: _containers.RepeatedCompositeFieldContainer[EnterpriseStatDetail]
    lastUpdated: int
    continuationToken: EnterpriseStatContinuationToken
    hasMore: bool
    def __init__(self, enterpriseStatDetails: _Optional[_Iterable[_Union[EnterpriseStatDetail, _Mapping]]] = ..., lastUpdated: _Optional[int] = ..., continuationToken: _Optional[_Union[EnterpriseStatContinuationToken, _Mapping]] = ..., hasMore: bool = ...) -> None: ...

class SecurityAlertsSummary(_message.Message):
    __slots__ = ["auditEventTypeId", "currentCount", "currentUserCount", "previousCount", "previousUserCount"]
    AUDITEVENTTYPEID_FIELD_NUMBER: _ClassVar[int]
    CURRENTCOUNT_FIELD_NUMBER: _ClassVar[int]
    CURRENTUSERCOUNT_FIELD_NUMBER: _ClassVar[int]
    PREVIOUSCOUNT_FIELD_NUMBER: _ClassVar[int]
    PREVIOUSUSERCOUNT_FIELD_NUMBER: _ClassVar[int]
    auditEventTypeId: int
    currentCount: int
    currentUserCount: int
    previousCount: int
    previousUserCount: int
    def __init__(self, auditEventTypeId: _Optional[int] = ..., currentCount: _Optional[int] = ..., currentUserCount: _Optional[int] = ..., previousCount: _Optional[int] = ..., previousUserCount: _Optional[int] = ...) -> None: ...

class SecurityAlertsSummaryResponse(_message.Message):
    __slots__ = ["securityAlertsSummary"]
    SECURITYALERTSSUMMARY_FIELD_NUMBER: _ClassVar[int]
    securityAlertsSummary: _containers.RepeatedCompositeFieldContainer[SecurityAlertsSummary]
    def __init__(self, securityAlertsSummary: _Optional[_Iterable[_Union[SecurityAlertsSummary, _Mapping]]] = ...) -> None: ...

class SecurityAlertsDetailRequest(_message.Message):
    __slots__ = ["auditEventTypeId", "continuationToken"]
    AUDITEVENTTYPEID_FIELD_NUMBER: _ClassVar[int]
    CONTINUATIONTOKEN_FIELD_NUMBER: _ClassVar[int]
    auditEventTypeId: int
    continuationToken: int
    def __init__(self, auditEventTypeId: _Optional[int] = ..., continuationToken: _Optional[int] = ...) -> None: ...

class SecurityAlertsDetail(_message.Message):
    __slots__ = ["enterpriseUserId", "currentCount", "previousCount", "lastOccurrence"]
    ENTERPRISEUSERID_FIELD_NUMBER: _ClassVar[int]
    CURRENTCOUNT_FIELD_NUMBER: _ClassVar[int]
    PREVIOUSCOUNT_FIELD_NUMBER: _ClassVar[int]
    LASTOCCURRENCE_FIELD_NUMBER: _ClassVar[int]
    enterpriseUserId: int
    currentCount: int
    previousCount: int
    lastOccurrence: int
    def __init__(self, enterpriseUserId: _Optional[int] = ..., currentCount: _Optional[int] = ..., previousCount: _Optional[int] = ..., lastOccurrence: _Optional[int] = ...) -> None: ...

class SecurityAlertsDetailResponse(_message.Message):
    __slots__ = ["securityAlertDetails", "hasMore", "continuationToken"]
    SECURITYALERTDETAILS_FIELD_NUMBER: _ClassVar[int]
    HASMORE_FIELD_NUMBER: _ClassVar[int]
    CONTINUATIONTOKEN_FIELD_NUMBER: _ClassVar[int]
    securityAlertDetails: _containers.RepeatedCompositeFieldContainer[SecurityAlertsDetail]
    hasMore: bool
    continuationToken: int
    def __init__(self, securityAlertDetails: _Optional[_Iterable[_Union[SecurityAlertsDetail, _Mapping]]] = ..., hasMore: bool = ..., continuationToken: _Optional[int] = ...) -> None: ...

class EnterpriseSecurityBenchmark(_message.Message):
    __slots__ = ["securityBenchmark", "securityBenchmarkStatus", "lastUpdated", "autoResolve"]
    SECURITYBENCHMARK_FIELD_NUMBER: _ClassVar[int]
    SECURITYBENCHMARKSTATUS_FIELD_NUMBER: _ClassVar[int]
    LASTUPDATED_FIELD_NUMBER: _ClassVar[int]
    AUTORESOLVE_FIELD_NUMBER: _ClassVar[int]
    securityBenchmark: SecurityBenchmark
    securityBenchmarkStatus: SecurityBenchmarkStatus
    lastUpdated: int
    autoResolve: bool
    def __init__(self, securityBenchmark: _Optional[_Union[SecurityBenchmark, str]] = ..., securityBenchmarkStatus: _Optional[_Union[SecurityBenchmarkStatus, str]] = ..., lastUpdated: _Optional[int] = ..., autoResolve: bool = ...) -> None: ...

class SetSecurityBenchmarksRequest(_message.Message):
    __slots__ = ["enterpriseSecurityBenchmarks"]
    ENTERPRISESECURITYBENCHMARKS_FIELD_NUMBER: _ClassVar[int]
    enterpriseSecurityBenchmarks: _containers.RepeatedCompositeFieldContainer[EnterpriseSecurityBenchmark]
    def __init__(self, enterpriseSecurityBenchmarks: _Optional[_Iterable[_Union[EnterpriseSecurityBenchmark, _Mapping]]] = ...) -> None: ...

class GetSecurityBenchmarksResponse(_message.Message):
    __slots__ = ["enterpriseSecurityBenchmarks"]
    ENTERPRISESECURITYBENCHMARKS_FIELD_NUMBER: _ClassVar[int]
    enterpriseSecurityBenchmarks: _containers.RepeatedCompositeFieldContainer[EnterpriseSecurityBenchmark]
    def __init__(self, enterpriseSecurityBenchmarks: _Optional[_Iterable[_Union[EnterpriseSecurityBenchmark, _Mapping]]] = ...) -> None: ...



from __future__ import annotations
import argparse
import logging
from . import PAMGatewayActionDiscoverCommandBase, GatewayContext
from ..pam.pam_dto import GatewayActionDiscoverJobRemoveInputs, GatewayActionDiscoverJobRemove, GatewayAction
from ..pam.router_helper import router_send_action_to_gateway, router_get_connected_gateways
from ... import vault_extensions
from ...display import bcolors
from ...discovery_common.jobs import Jobs
from ...proto import pam_pb2


class PAMGatewayActionDiscoverJobRemoveCommand(PAMGatewayActionDiscoverCommandBase):
    parser = argparse.ArgumentParser(prog='pam-action-discover-remove')
    parser.add_argument('--job-id', '-j', required=True, dest='job_id', action='store',
                        help='Discovery job id.')

    def get_parser(self):
        return PAMGatewayActionDiscoverJobRemoveCommand.parser

    def execute(self, params, **kwargs):

        if not hasattr(params, 'pam_controllers'):
            router_get_connected_gateways(params)

        job_id = kwargs.get("job_id")

        # Get all the PAM configuration records
        configuration_records = list(vault_extensions.find_records(params, "pam.*Configuration"))
        all_gateways = GatewayContext.all_gateways(params)

        for configuration_record in configuration_records:

            gateway_context = GatewayContext.from_configuration_uid(params=params,
                                                                    configuration_uid=configuration_record.record_uid,
                                                                    gateways=all_gateways)
            if gateway_context is None:
                continue

            jobs = Jobs(record=configuration_record, params=params)
            job_item = jobs.get_job(job_id)
            if job_item is not None:

                try:
                    # First, cancel the running discovery job if it is running.
                    logging.debug("cancel job on the gateway, if running")
                    action_inputs = GatewayActionDiscoverJobRemoveInputs(
                        configuration_uid=gateway_context.configuration_uid,
                        job_id=job_id
                    )

                    conversation_id = GatewayAction.generate_conversation_id()
                    router_response = router_send_action_to_gateway(
                        params=params,
                        gateway_action=GatewayActionDiscoverJobRemove(
                            inputs=action_inputs,
                            conversation_id=conversation_id),
                        message_type=pam_pb2.CMT_DISCOVERY,
                        is_streaming=False,
                        destination_gateway_uid_str=gateway_context.gateway_uid
                    )

                    data = self.get_response_data(router_response)
                    if data is None:
                        raise Exception("The router returned a failure.")
                    elif data.get("success") is False:
                        error = data.get("error")
                        raise Exception(f"Discovery job was not removed: {error}")
                except Exception as err:
                    logging.debug(f"gateway return error removing discovery job: {err}")

                jobs.cancel(job_id)

                print(f"{bcolors.OKGREEN}Discovery job has been removed or cancelled.{bcolors.ENDC}")
                return

        print(f'{bcolors.FAIL}Discovery job not found. Cannot get remove the job.{bcolors.ENDC}')
        return



import sys
if sys.version_info >= (3, 8):
    import pytest
    from unittest import TestCase, mock
    from flask import Flask, request
    from keepercommander.service.decorators.api_logging import api_log_handler


    class TestApiLogging(TestCase):
        def setUp(self):
            self.app = Flask(__name__)
            self.client = self.app.test_client()

            @self.app.route('/test', methods=['POST'])
            @api_log_handler
            def test_endpoint():
                if not request.is_json:
                    return {'error': 'Content-Type must be application/json'}, 415
                return {'status': 'success'}, 200

            @self.app.route('/error', methods=['POST'])
            @api_log_handler
            def error_endpoint():
                if not request.is_json:
                    return {'error': 'Content-Type must be application/json'}, 415
                raise Exception("Test error")

        def test_api_log_success_request(self):
            """Test logging of successful API request"""
            with mock.patch('keepercommander.service.decorators.api_logging.logger.info') as mock_log:
                test_data = {"test": "data"}
                response = self.client.post('/test',
                                        json=test_data,
                                        headers={
                                            'X-Forwarded-For': '127.0.0.1',
                                            'Content-Type': 'application/json'
                                        })

                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.json, {'status': 'success'})

                mock_log.assert_called_once()
                log_message = mock_log.call_args[0][0]
                self.assertIn('POST', log_message)
                self.assertIn('/test', log_message)
                self.assertIn('127.0.0.1', log_message)
                self.assertIn('200', log_message)
                self.assertIn(f"data={str(test_data)}", log_message)

        def test_api_log_error_request(self):
            """Test logging of failed API request"""
            with mock.patch('keepercommander.service.decorators.api_logging.logger.error') as mock_log:
                response = self.client.post('/error', json={}, 
                                        headers={'X-Forwarded-For': '127.0.0.1', 
                                                'Content-Type': 'application/json'})

                self.assertEqual(response.status_code, 500)
                mock_log.assert_called_once()
                log_message = mock_log.call_args[0][0]
                self.assertIn('POST', log_message)
                self.assertIn('/error', log_message)
                self.assertIn('127.0.0.1', log_message)
                self.assertIn("error='Test error'", log_message)

        def test_api_log_remote_addr_fallback(self):
            """Test logging falls back to remote_addr when X-Forwarded-For is missing"""
            with mock.patch('keepercommander.service.decorators.api_logging.logger.info') as mock_log:
                response = self.client.post('/test', json={}, 
                                        headers={'Content-Type': 'application/json'}, 
                                        environ_base={'REMOTE_ADDR': '192.168.1.1'})

                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.json, {'status': 'success'})

                mock_log.assert_called_once()
                log_message = mock_log.call_args[0][0]
                self.assertIn('192.168.1.1', log_message)

        def test_api_log_timing(self):
            """Test request timing is logged"""
            with mock.patch('keepercommander.service.decorators.api_logging.logger.info') as mock_log:
                response = self.client.post('/test', json={}, 
                                        headers={'Content-Type': 'application/json'})

                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.json, {'status': 'success'})

                mock_log.assert_called_once()
                log_message = mock_log.call_args[0][0]
                self.assertRegex(log_message, r'\d+\.\d+s')



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2025 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse

from .base import BiometricCommand
from ...commands.base import user_choice
from ..utils.constants import get_status_message


class BiometricUpdateNameCommand(BiometricCommand):
    """Update friendly name of a biometric passkey"""

    parser = argparse.ArgumentParser(prog='biometric update-name', description='Update friendly name of a biometric passkey')

    def get_parser(self):
        return self.parser

    def execute(self, params, **kwargs):
        """Execute biometric update-name command"""
        def _update_name():
            all_credentials = self._get_available_credentials_or_error(params)
            available_credentials = [cred for cred in all_credentials if cred.get('name') and cred['name'].strip()]
            
            if not available_credentials:
                print("No biometric credentials with friendly names found.")
                print("Use 'biometric register' to create a new credential with a friendly name.")
                return
                
            print(f"Found {len(available_credentials)} biometric credential(s) with friendly names")

            selected_index = self._interactive_credential_selection(available_credentials)

            target_credential = available_credentials[selected_index]

            friendly_name = self._interactive_name_input(target_credential)

            if len(friendly_name) > 32:
                raise Exception("Friendly name must be 32 characters or less")

            if not self._confirm_update(target_credential, friendly_name):
                print("Update cancelled by user")
                return

            result = self.client.update_passkey_name(params, target_credential['id'], target_credential['credential_id'], friendly_name)
            
            self._report_update_results(result, target_credential, friendly_name)

        return self._execute_with_error_handling('update passkey friendly name', _update_name)

    def _interactive_credential_selection(self, available_credentials):
        """Interactive selection of credential to update"""
        if len(available_credentials) == 1:
            credential = available_credentials[0]
            print(f"Found single credential: {credential['name']}")
            answer = user_choice('Use this credential?', 'yn', 'y')
            if answer.lower() == 'y':
                return 0  # Return index instead of ID
            else:
                raise Exception("Operation cancelled by user")
        
        print("\nAvailable Biometric Credentials:")
        print("-" * 50)
        
        for i, credential in enumerate(available_credentials, 1):
            created_date = self._format_timestamp(credential.get('created'))
            last_used = self._format_timestamp(credential.get('last_used'))
            
            print(f"{i:2}. {credential['name']}")
            print(f"    Created: {created_date}")
            print(f"    Last Used: {last_used}")
            print()
        
        while True:
            try:
                choice = input(f"Select credential number (1-{len(available_credentials)}): ")
                if choice.lower() in ['q', 'quit', 'exit']:
                    raise Exception("Operation cancelled by user")
                
                selection = int(choice) - 1
                if 0 <= selection < len(available_credentials):
                    selected_credential = available_credentials[selection]
                    print(f"Selected: {selected_credential['name']}")
                    return selection  # Return index instead of ID
                else:
                    print(f"Invalid selection. Please choose 1-{len(available_credentials)}.")
            except ValueError:
                print("Invalid input. Please enter a number.")
            except KeyboardInterrupt:
                raise Exception("Operation cancelled by user")

    def _interactive_name_input(self, credential):
        """Interactive input for new friendly name"""
        print(f"\nCurrent Name: {credential['name']}")
        print("Enter a new friendly name (max 32 characters):")
        
        while True:
            try:
                new_name = input("New name: ").strip()
                
                if not new_name:
                    print("Name cannot be empty. Please try again.")
                    continue
                
                if len(new_name) > 32:
                    print(f"Name too long ({len(new_name)} chars). Maximum 32 characters allowed.")
                    continue
                
                if new_name == credential['name']:
                    print("Name is the same as current name. Please enter a different name.")
                    continue
                
                return new_name
                
            except KeyboardInterrupt:
                raise Exception("Operation cancelled by user")

    def _confirm_update(self, credential, new_name):
        """Confirm the update operation"""
        print("\nUpdate Summary:")
        print("-" * 20)
        print(f"Current Name:  {credential['name']}")
        print(f"New Name:      {new_name}")
        print()
        
        answer = user_choice('Proceed with update?', 'yn', 'y')
        return answer.lower() == 'y'

    def _report_update_results(self, result, credential, new_name):
        """Report the update results to the user"""
        print("\nPasskey Update Results:")
        print("=" * 30)
        status_code = result['status']
        status_text = get_status_message(status_code)
        print(f"Status: {status_text}")
        print(f"Old Name: {credential['name']}")
        print(f"New Name: {new_name}")
        print(f"Message: {result['message']}")
        print("=" * 30) 


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: commander@keepersecurity.com
#
# Example script to run a BreachWatch status report, parse the results,
# and send users an email reminder to address their found issues.
#
# Note: SMTP credentials must be supplied via a vault record
# in order to send the email. 
# 
# This example also pulls configuration
# from config.json or writes the config file if it does not exist.
#
# Usage:
#    python send_breachwatch_reminder.py

import getpass
import json
import os
import ssl
from smtplib import SMTP

from keepercommander import api, vault_extensions, vault
from keepercommander.__main__ import get_params_from_config
from keepercommander.commands.security_audit import SecurityAuditReportCommand

email_message = '''
From: {0}
Subject: Keeper BreachWatch Alert

BreachWatch detected records at risk in your vault.
Please login to Keeper and review the records marked "At Risk".
'''

my_params = get_params_from_config(os.path.join(os.path.dirname(__file__), 'config.json'))

while not my_params.user:
    my_params.user = getpass.getpass(prompt='User(Email): ', stream=None)

api.login(my_params)

report_command = SecurityAuditReportCommand()
report_json = report_command.execute(my_params, breachwatch=True, format='json')
report = json.loads(report_json)
emails = [x['email'] for x in report if x.get('at_risk') > 5]
if emails:
    api.sync_down(my_params)
    smtp_record = next(vault_extensions.find_records(my_params, search_str='smtp', record_type='serverCredentials'), None)
    if isinstance(smtp_record, vault.TypedRecord):
        smtp_host = None
        smtp_port = 0
        username = None
        password = None
        field = smtp_record.get_typed_field('host')
        if field:
            host_value = field.get_default_value()
            if isinstance(host_value, dict):
                smtp_host = host_value.get('hostName')
                port = host_value.get('port')
                if port:
                    try:
                        smtp_port = int(port)
                    except ValueError:
                        pass
        if smtp_host:
            field = smtp_record.get_typed_field('login')
            if field:
                username = field.get_default_value()
            field = smtp_record.get_typed_field('password')
            if field:
                password = field.get_default_value()

        if smtp_host:
            with SMTP(host=smtp_host, port=smtp_port) as connection:
                if username:
                    connection.starttls(context=ssl.create_default_context())
                    connection.login(user=username, password=password)
                connection.sendmail(my_params.user, emails, email_message.format(my_params.user))



from __future__ import annotations
from typing import Any, Optional


class DAGException(Exception):

    def __init__(self, msg: Any, uid: Optional[str] = None):
        if isinstance(msg, str) is False:
            msg = str(msg)

        self.msg = msg
        self.uid = uid

        super().__init__(self.msg)

    def __str__(self):
        return self.msg

    def __repr__(self):
        return self.msg


class DAGKeyIsEncryptedException(DAGException):
    pass


class DAGDataEdgeNotFoundException(DAGException):
    pass


class DAGDeletionException(DAGException):
    pass


class DAGConfirmException(DAGException):
    pass


class DAGPathException(DAGException):
    pass


class DAGVertexAlreadyExistsException(DAGException):
    pass


class DAGContentException(DAGException):
    pass


class DAGDefaultGraphException(DAGException):
    pass


class DAGIllegalEdgeException(DAGException):
    pass


class DAGKeyException(DAGException):
    pass


class DAGDataException(DAGException):
    pass


class DAGVertexException(DAGException):
    pass


class DAGEdgeException(DAGException):
    pass


class DAGCorruptException(DAGException):
    pass


class DAGConnectionException(DAGException):
    pass



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2019 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#
# Example showing how to access the raw enterprise event logs
# for some custom processing.
#

from keepercommander import api
from keepercommander.__main__ import get_params_from_config

my_params = get_params_from_config('config.json')
while not my_params.user:
    my_params.user = input('User(Email): ')
api.login(my_params)
if not my_params.session_token:
    exit(1)

events = []
finished = False
# UNIX epoch time in seconds
last_event_time = 0
logged_ids = set()

print('Downloading ', end='', flush=True)

while not finished:
    finished = True
    rq = {
        'command': 'get_audit_event_reports',
        'report_type': 'raw',
        'scope': 'enterprise',
        'limit': 1000,
        'order': 'ascending'
    }

    if last_event_time > 0:
        rq['filter'] = {
            'created': {'min': last_event_time}  # return audit events starting last_event_time
        }

    rs = api.communicate(my_params, rq)
    print('.', end='', flush=True)
    if rs['result'] == 'success':
        finished = True
        if 'audit_event_overview_report_rows' in rs:
            audit_events = rs['audit_event_overview_report_rows']
            if len(audit_events) > 1:
                # remove events from the tail for the last second
                last_event_time = int(audit_events[-1]['created'])
                while len(audit_events) > 0:
                    event = audit_events[-1]
                    if int(event['created']) < last_event_time:
                        break
                    audit_events = audit_events[:-1]

                for event in audit_events:
                    event_id = event['id']
                    if event_id not in logged_ids:
                        logged_ids.add(event_id)
                        events.append(event)
                finished = len(events) == 0

print()
print("{0} audit log events".format(len(events)))



import logging
from typing import Iterable, Dict, Set, List, Optional

from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKey

from . import sox_types, sqlite_storage, storage_types
from .sox_types import RecordPermissions, SharedFolder
from .. import crypto, utils
from ..error import Error
from ..params import KeeperParams


class RebuildTask:
    def __init__(self, is_full_sync, load_compliance_data=False, load_aging_data=False):
        # type: (bool, bool, bool) -> None
        self.is_full_sync = is_full_sync                    # type: bool
        self.load_compliance_data = load_compliance_data    # type: bool
        self.records = set()                                # type: Set[str]
        self.load_aging_data = load_aging_data              # type: bool

    def update_records(self, record_ids):    # type: (Iterable[str]) -> None
        if self.is_full_sync:
            return
        self.records.update(record_ids)


def get_ec_private_key(params):  # type: (KeeperParams) -> EllipticCurvePrivateKey
    tree_key = params.enterprise['unencrypted_tree_key']
    ecc_key = utils.base64_url_decode(params.enterprise['keys']['ecc_encrypted_private_key'])
    ecc_key = crypto.decrypt_aes_v2(ecc_key, tree_key)
    return crypto.load_ec_private_key(ecc_key)


def clear_lookup(lookup, uids=None):  # type: (dict, Optional[Iterable]) -> None
    if uids:
        [lookup.pop(k) for k in uids]
    else:
        lookup.clear()


class SoxData:
    def __init__(self, params, storage, no_cache=False):
        # type: (KeeperParams, sqlite_storage.SqliteSoxStorage, Optional[bool]) -> None
        self.storage = storage                  # type: sqlite_storage.SqliteSoxStorage
        self._records = {}                      # type: Dict[str, sox_types.Record]
        self._users = {}                        # type: Dict[int, sox_types.EnterpriseUser]
        self._teams = {}                        # type: Dict[str, sox_types.Team]
        self._shared_folders = {}               # type: Dict[str, sox_types.SharedFolder]
        self.ec_private_key = get_ec_private_key(params)
        self.tree_key = params.enterprise.get('unencrypted_tree_key', b'')
        task = RebuildTask(True)
        self.rebuild_data(task, no_cache)

    def get_records(self, record_ids=None):
        return self._records if record_ids is None else {uid: self._records.get(uid) for uid in record_ids}

    def get_user(self, uid):    # type: (int) -> sox_types.EnterpriseUser
        return self._users.get(uid)

    def get_users(self, user_ids=None):
        return self._users if user_ids is None else {uid: self.get_user(uid) for uid in user_ids}

    def get_team(self, team_uid):
        return self._teams.get(team_uid)

    def get_teams(self, team_uids=None):
        return self._teams if team_uids is None else {uid: self.get_team(uid) for uid in team_uids}

    def get_user_records(self, user_uids=None):
        if user_uids:
            users = set()
            for uid in user_uids:
                user = self.get_user(uid)
                if user:
                    users.add(user)
        else:
            users = self._users.values()

        recs = set()
        for user in users:
            for r_uid in user.records:
                recs.add(sox_types.UserRecord(user.user_uid, self._records.get(r_uid)))
        return recs

    def get_shared_folders(self, sf_uids=None):
        return self._shared_folders if sf_uids is None else {uid: self._shared_folders.get(uid) for uid in sf_uids}

    def get_record_sfs(self, record_uid):
        get_ruids = lambda sf: [rp.record_uid for rp in sf.record_permissions]
        return [sf.folder_uid for sf in self._shared_folders.values() if record_uid in get_ruids(sf)]

    def get_record_owner(self, rec_uid):
        owner = None
        if rec_uid in self._records:
            for user in self._users.values():
                if rec_uid in user.records:
                    owner = user
                    break
        return owner

    def get_vault_records(self, user_ref):
        user_id = user_ref if isinstance(user_ref, int) or user_ref.isdigit() else next((k for k, v in self._users.items() if v.email == user_ref), 0)
        owned_sa_ruids = {ruid: rec for ruid, rec in self._records.items() if user_id in rec.user_permissions.keys()}

        # Get SF-shared records
        # Shared to team
        user_teams = {tuid for tuid, t in self._teams.items() if user_id in t.users}
        team_sf_uids = {sfuid for sfuid, sf in self._shared_folders.items() if sf.teams.intersection(user_teams)}
        # Shared to user
        user_sf_uids = {sfuid for sfuid, sf in self._shared_folders.items() if user_id in sf.users}
        sf_uids = team_sf_uids.union(user_sf_uids)
        user_sfs = [sf for uid, sf in self._shared_folders.items() if uid in sf_uids]
        user_sf_rec_uids = {rp.record_uid for sf in user_sfs for rp in sf.record_permissions}

        vault_record_uids = user_sf_rec_uids.union(owned_sa_ruids)
        return self.get_records(vault_record_uids)

    def clear_records(self, uids=None):
        clear_lookup(self._records, uids)

    def clear_users(self, uids=None):
        clear_lookup(self._users, uids)

    def clear_teams(self, uids=None):
        clear_lookup(self._teams, uids)

    def clear_shared_folders(self, uids=None):
        clear_lookup(self._shared_folders, uids)

    def clear_all(self):
        self.clear_records()
        self.clear_users()
        self.clear_teams()
        self.clear_shared_folders()

    @property
    def record_count(self):   # type: () -> int
        return len(self._records)

    def rebuild_data(self, changes, no_cache=False):   # type: (RebuildTask, Optional[bool]) -> None
        def decrypt(data):  # type: (bytes or str) -> str
            if isinstance(data, str):
                return data

            decrypted = ''
            try:
                decrypted_bytes = crypto.decrypt_aes_v1(data, self.tree_key) if data else b''
                decrypted = decrypted_bytes.decode()
            except Error as e:
                logging.info(f'Error decrypting data: type = {type(data)}, value = {data}, message = {e.message}')
            return decrypted

        def link_record_permissions(store, record_lookup):
            links = store.get_record_permissions().get_all_links()
            for link in links:
                record = record_lookup.get(link.record_uid)
                if record:
                    record.user_permissions.update({link.user_uid: link.permissions})
                else:
                    logging.info(f'record (uid = {link.record_uid}) not found')

            return record_lookup

        def load_records(store, changes):
            # type: (sqlite_storage.SqliteSoxStorage, RebuildTask) -> Dict[str, sox_types.Record]
            entities = []   # type: List[storage_types.StorageRecord]
            if changes.records:
                for uid in changes.records:
                    entity = store.records.get_entity(uid)
                    if entity:
                        entities.append(entity)
            else:
                entities.extend(store.records.get_all())

            record_lookup = {}
            for entity in entities:
                record = self._records.get(entity.record_uid) or sox_types.Record()
                record.update_properties(entity, self.ec_private_key)
                record_lookup[record.record_uid] = record

            record_lookup = link_record_aging(store, record_lookup)
            return link_record_permissions(store, record_lookup) if changes.load_compliance_data else record_lookup

        def link_record_aging(store, record_lookup):
            for aging in store.record_aging.get_all():
                record = record_lookup.get(aging.record_uid)
                if record:
                    record.created = aging.created
                    record.last_pw_change = aging.last_pw_change
                    record.last_modified = aging.last_modified
                    record.last_rotation = aging.last_rotation
            return record_lookup

        def link_user_records(store, user_lookup):
            links = store.get_user_record_links().get_all_links()
            for link in links:
                user = user_lookup.get(link.user_uid)
                rec = self._records.get(link.record_uid)
                if user:
                    user.trash_records.add(rec.record_uid) if rec.in_trash else user.active_records.add(rec.record_uid)
                    user.records.add(link.record_uid)
                else:
                    logging.info(f'user (uid = {link.user_uid} not found')
            return user_lookup

        def load_users(store):  # type: (sqlite_storage.SqliteSoxStorage) -> Dict[int, sox_types.EnterpriseUser]
            users = [sox_types.EnterpriseUser.load(eu, decrypt_fn=decrypt) for eu in store.users.get_all()]
            u_lookup = {user.user_uid: user for user in users}
            return link_user_records(store, u_lookup)

        def load_teams(store):
            teams = [sox_types.Team.load(entity) for entity in store.teams.get_all()]
            team_lookup = {team.team_uid: team for team in teams}
            return link_team_users(store, team_lookup)

        def link_team_users(store, team_lookup):
            links = store.get_team_user_links().get_all_links()
            for link in links:
                team = team_lookup.get(link.team_uid)
                team.users.append(link.user_uid)
            return team_lookup

        def load_shared_folders(store):
            sf_lookup = dict()  # type: Dict[str, SharedFolder]
            sf_lookup = link_sf_records(store, sf_lookup)
            sf_lookup = link_sf_users(store, sf_lookup)
            sf_lookup = link_sf_teams(store, sf_lookup)
            return sf_lookup

        def link_sf_records(store, folder_lookup):
            sf_record_links = store.get_sf_record_links().get_all_links()
            for link in sf_record_links:
                folder = folder_lookup[link.folder_uid] if link.folder_uid in folder_lookup \
                    else SharedFolder(link.folder_uid)
                folder.update_record_permissions(RecordPermissions(link.record_uid, link.permissions))
                folder_lookup[link.folder_uid] = folder

            return folder_lookup

        def link_sf_users(store, folder_lookup):
            sf_user_links = store.get_sf_user_links().get_all_links()
            for link in sf_user_links:
                folder = folder_lookup[link.folder_uid] if link.folder_uid in folder_lookup \
                    else SharedFolder(link.folder_uid)
                folder.users.add(link.user_uid)
                folder_lookup[link.folder_uid] = folder

            return folder_lookup

        def link_sf_teams(store, folder_lookup):
            sf_team_links = store.get_sf_team_links().get_all_links()
            for link in sf_team_links:
                folder = folder_lookup[link.folder_uid] if link.folder_uid in folder_lookup \
                    else SharedFolder(link.folder_uid)
                folder.teams.add(link.team_uid)
                folder_lookup[link.folder_uid] = folder

            return folder_lookup

        if changes.is_full_sync:
            self.clear_all()
        if changes.load_aging_data:
            self.clear_records(changes.records)
        if changes.load_compliance_data:
            self._teams.update(load_teams(self.storage))
            self._shared_folders.update(load_shared_folders(self.storage))
        self._records.update(load_records(self.storage, changes))
        if changes.is_full_sync or changes.load_compliance_data:
            self._users.update(load_users(self.storage))
        if no_cache:
            self.storage.delete_db()
